/*
    Brightness in 10Lux
*/

#define LIGHT_TORCH vec3(1.0, 0.55, 0.18)
#define LIGHT_REDTORCH vec3(1.0, 0.3, 0.1)
#define LIGHT_SOUL vec3(0.26, 0.42, 1.0)
#define LIGHT_FIRE vec3(1.0, 0.45, 0.12)
#define LIGHT_ENDROD vec3(0.76, 0.35, 1.0)
#define LIGHT_AMETHYST vec3(0.55, 0.25, 1.0)

vec3 getEmission(int ID, vec3 albedo) {
    albedo  = max(albedo, 1e-8);
    float lum   = getLuma(albedo);
    //albedo  = saturate(mix(cube(lum) * normalize(albedo), albedo, lum));

    float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
        albedoLum   = saturate(albedoLum * sqrt2);

    float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

    albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;

    switch (ID) {
        case 1: return albedo * 200;
        case 2: return albedo * 160;
        case 3: return albedo * 80;
        case 4: return albedo * 10;
        case 10: return LIGHT_TORCH * 120;
        case 11: return LIGHT_REDTORCH * 35;
        case 12: return LIGHT_SOUL * 60;
        case 13: return LIGHT_FIRE * 80;
        case 20: return LIGHT_ENDROD * 100;
        case 21: return LIGHT_AMETHYST * 20;
        case 22: return LIGHT_AMETHYST * 5;
    }

    return vec3(0.0);
}

vec3 getEmissionScreenSpace(int ID, vec3 albedo, float textureEmission) {
    albedo  = max(albedo, 1e-8);
    vec3 albedoIn   = albedo;

    #if labEmissionMode != 2
        float lum   = getLuma(albedo);

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

        #if labEmissionMode == 1
            emitterLum  = max(emitterLum, textureEmission);
        #endif

        albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;
    #else
        albedo   *= textureEmission;
        float emitterLum = textureEmission;
    #endif

    switch (ID) {
        case 1: return albedo * 200;
        case 2: return albedo * 160;
        case 3: return albedo * 80;
        case 4: return albedo * 10;
        case 10: return LIGHT_TORCH * 120 * emitterLum;
        case 11: return LIGHT_REDTORCH * 35 * emitterLum;
        case 12: return LIGHT_SOUL * 60 * emitterLum;
        case 13: return LIGHT_FIRE * 80 * emitterLum;
        case 20: return LIGHT_ENDROD * 100 * emitterLum;
        case 21: return LIGHT_AMETHYST * 20;
        case 22: return LIGHT_AMETHYST * 5;
    }

    #if labEmissionMode == 0
        return vec3(0);
    #elif labEmissionMode == 1
        return albedoIn * textureEmission * 80;
    #elif labEmissionMode == 2
        return albedo * 80;
    #endif
}