/*
    Voxel Structure based on Superscalar by Bobcao3
*/

uniform float eyeAltitude;

const int volumeWidth = 128;
const int volumeHeight = 128;

const int voxelmapResolution = voxelMapRes;
const ivec2 voxelSegmentEdge = ivec2(voxelmapResolution, voxelmapResolution / 2) - 1;

const int volumeDepthGridSize = (voxelmapResolution / volumeWidth);

const int volumeDepth = ((voxelmapResolution / volumeWidth) * (voxelmapResolution / volumeHeight)) / 2;

vec3 sceneToVoxelSpace(vec3 scenePos) {
    return scenePos + mod(cameraPosition, 1.0) + vec3(volumeWidth, volumeDepth, volumeHeight) / 2;
}
ivec3 sceneToVoxelSpaceInt(vec3 scenePos) {
    return ivec3(floor(scenePos + mod(cameraPosition, 1.0))) + ivec3(volumeWidth, volumeDepth, volumeHeight) / 2;
}
ivec3 sceneToVoxelSpaceNoCam(vec3 scenePos) {
    return ivec3(floor(scenePos)) + ivec3(volumeWidth, volumeDepth, volumeHeight) / 2;
}
vec3 voxelToSceneSpace(ivec3 voxelPos) {
    return vec3(voxelPos) - vec3(volumeWidth, volumeDepth, volumeHeight) / 2.0;
}
vec3 voxelToSceneSpace(vec3 voxelPos) {
    return vec3(voxelPos) - vec3(volumeWidth, volumeDepth, volumeHeight) / 2.0;
}

bool outsideVoxelVolume(ivec3 pos) {
    return pos.x < 0 || pos.x >= volumeWidth || pos.y < 0 || pos.y >= volumeDepth || pos.z < 0 || pos.z >= volumeHeight;
}
bool outsideVoxelVolume(vec3 pos) {
    return pos.x < 0 || pos.x >= volumeWidth || pos.y < 0 || pos.y >= volumeDepth || pos.z < 0 || pos.z >= volumeHeight;
}
bool sceneWithinVoxelVolume(vec3 pos) {
    return !outsideVoxelVolume(sceneToVoxelSpaceNoCam(pos));
}
bool withinVoxelVolume(ivec3 index) {
    return !outsideVoxelVolume(index);
}
bool withinVoxelVolume(vec3 index) {
    return !outsideVoxelVolume(index);
}


bool outsideVoxelVolume(ivec3 pos, int lod) {
    int lodFactor   = int(pow(2, lod));

    ivec3 bounds    = ivec3(volumeWidth, volumeDepth, volumeHeight) / lodFactor;
    return pos.x < 0 || pos.x >= bounds.x || pos.y < 0 || pos.y >= bounds.y || pos.z < 0 || pos.z >= bounds.z;
}
bool outsideVoxelVolume(vec3 pos, int lod) {
    int lodFactor   = int(pow(2, lod));

    ivec3 bounds    = ivec3(volumeWidth, volumeDepth, volumeHeight) / lodFactor;
    return pos.x < 0 || pos.x >= bounds.x || pos.y < 0 || pos.y >= bounds.y || pos.z < 0 || pos.z >= bounds.z;
}
bool withinVoxelVolume(ivec3 index, int lod) {
    return !outsideVoxelVolume(index, lod);
}
bool withinVoxelVolume(vec3 index, int lod) {
    return !outsideVoxelVolume(index, lod);
}

ivec2 getVoxelIndex(ivec3 pos) {
    if (outsideVoxelVolume(pos)) return ivec2(-1);

    ivec2 base = ivec2((pos.y % volumeDepthGridSize) * volumeWidth, (pos.y / volumeDepthGridSize) * volumeHeight);
    ivec2 UV = base + pos.xz;

    return UV;
}
vec2 getVoxelCoord(vec3 pos) {
    if (outsideVoxelVolume(pos)) return ivec2(-1);

    ivec2 base = ivec2((int(pos.y) % volumeDepthGridSize) * volumeWidth, (int(pos.y) / volumeDepthGridSize) * volumeHeight);
    vec2 uv = vec2(base) + pos.xz;

    return (uv + 0.5) / voxelmapResolution;
}
ivec3 uvToVoxelIndex(ivec2 uv) {
    ivec2 y_grid = uv / ivec2(volumeWidth, volumeHeight);
    ivec2 xz_pos = uv % ivec2(volumeWidth, volumeHeight);
    ivec3 wpos = ivec3(xz_pos.x, y_grid.x + y_grid.y * volumeDepthGridSize, xz_pos.y);

    return wpos;
}

ivec2 getVoxelIndex(ivec3 pos, int lod) {
    if (outsideVoxelVolume(pos)) return ivec2(-1);

    int lodFactor   = int(pow(2, lod));

    ivec3 bounds    = ivec3(volumeWidth, volumeDepth, volumeHeight) / lodFactor;
    int volumeDepthGridSize = (voxelmapResolution / lodFactor) / bounds.x;

    ivec2 base = ivec2((pos.y % volumeDepthGridSize) * bounds.x, (pos.y / volumeDepthGridSize) * bounds.z);
    ivec2 UV = base + pos.xz;

    return UV;
}
vec2 getVoxelCoord(vec3 pos, int lod) {
    if (outsideVoxelVolume(pos)) return ivec2(-1);

    int lodFactor   = int(pow(2, lod));

    ivec3 bounds    = ivec3(volumeWidth, volumeDepth, volumeHeight) / lodFactor;
    int volumeDepthGridSize = (voxelmapResolution) / bounds.x;

    ivec2 base = ivec2(mod(pos.y, volumeDepthGridSize) * bounds.x, (pos.y / volumeDepthGridSize) * bounds.z);
    vec2 uv = vec2(base) + pos.xz;

    return (uv + 0.5) / voxelmapResolution;
}
ivec3 uvToVoxelIndex(ivec2 uv, int lod) {
    int lodFactor   = int(pow(2, lod));

    ivec3 bounds    = ivec3(volumeWidth, volumeDepth, volumeHeight) / lodFactor;
    int volumeDepthGridSize = (voxelmapResolution) / bounds.x;

    ivec2 y_grid = uv / ivec2(bounds.x, bounds.z);
    ivec2 xz_pos = uv % ivec2(bounds.x, bounds.z);
    ivec3 wpos = ivec3(xz_pos.x, y_grid.x + y_grid.y * volumeDepthGridSize, xz_pos.y);

    return wpos;
}
ivec3 uvToVoxelIndex1(vec2 uv, int lod) {
    int lodFactor   = int(pow(2, lod));

    ivec3 bounds    = ivec3(volumeWidth, volumeDepth, volumeHeight) / lodFactor;
    int volumeDepthGridSize = (voxelmapResolution / lodFactor) / bounds.x;

    ivec2 UV    = ivec2(floor(uv * voxelmapResolution / lodFactor));

    ivec2 y_grid = UV / ivec2(bounds.x, bounds.z);
    ivec2 xz_pos = UV % ivec2(bounds.x, bounds.z);
    ivec3 wpos = ivec3(xz_pos.x, y_grid.x + y_grid.y * volumeDepthGridSize, xz_pos.y);

    return wpos;
}