/* RENDERTARGETS: 15 */
layout(location = 0) out vec4 cacheOut;


#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex7;
uniform sampler2D colortex15;

const bool shadowHardwareFiltering = true;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform sampler2D noisetex;

uniform int frameCounter;

uniform float frameTime, frameTimeCounter;
uniform float far, near;

uniform float sunAngle;

uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

in vec2 coord;

#include "/lib/util/colorspace.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/voxel/store.glsl"

#include "/lib/light/cacheLib.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};


void main() {
    ivec2 pixelUV   = ivec2(gl_FragCoord.xy);

    vec4 cacheResult = texture(colortex15, coord);

    int face    = getFaceFromUV(pixelUV);

    if (face > 7) {
        ivec2 facePixel = pixelUV & cacheSegmentEdge;
        vec2 uv         = (vec2(facePixel) + 0.5) / cacheSegmentResolution;
        ivec3 index     = uvToCacheIndex(facePixel);

        vec4 sumAllFaces    = vec4(0);
        for (int i = 0; i < 6; ++i) {
            sumAllFaces    += cacheRead(getCachePixelCacheSpace(index, i));
        }
        sumAllFaces        /= 6.0;

        cacheResult         = sumAllFaces;
    }

    cacheOut    = clamp16F(cacheResult);
}