/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 6,7,10,11,12,13 */
layout(location = 0) out vec4 currentSH;
layout(location = 1) out vec4 currentCC;
layout(location = 2) out vec4 GbufferHistory;
layout(location = 3) out vec4 historySH;
layout(location = 4) out vec4 historyCC;
layout(location = 5) out vec4 historyAux;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

uniform sampler2D colortex1;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex10;
uniform sampler2D colortex11;
uniform sampler2D colortex12;
uniform sampler2D colortex13;

uniform sampler2D depthtex0;
uniform sampler2D depthtex2;

/*
const bool colortex10Clear  = false;
const bool colortex11Clear  = false;
const bool colortex12Clear  = false;
const bool colortex13Clear  = false;
*/

uniform float far, near;

uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousProjection, gbufferPreviousModelView;

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/util/transforms.glsl"

/* ------ reprojection ----- */
vec3 reproject(vec3 sceneSpace, bool hand) {
    vec3 prevScreenPos = hand ? vec3(0.0) : cameraPosition - previousCameraPosition;
    prevScreenPos = sceneSpace + prevScreenPos;
    prevScreenPos = transMAD(gbufferPreviousModelView, prevScreenPos);
    prevScreenPos = transMAD(gbufferPreviousProjection, prevScreenPos) * (0.5 / -prevScreenPos.z) + 0.5;

    return prevScreenPos;
}

/* ------ ATROUS ------ */

ivec2 clampUV(ivec2 UV) {
    return clamp(UV, ivec2(0.0), ivec2(viewSize - 1));
}
ivec2 clampUV_LR(ivec2 UV) {
    return clamp(UV, ivec2(0.0), ivec2(viewSize * indirectResScale - 1));
}
bool UV_WithinBounds(ivec2 UV) {
    return clamp(UV, ivec2(0), ivec2(viewSize * indirectResScale - 1)) == UV;
}

vec4 FetchUnpackedGbuffer(ivec2 UV) {
    vec4 Val  = texelFetch(colortex5, UV, 0);
    return vec4(Val.rgb * 2.0 - 1.0, sqr(Val.w));
}
vec4 FetchUnpackedGbufferRaw(ivec2 UV) {
    return texelFetch(colortex5, UV, 0);
}

float FetchHistoryDistance(ivec2 UV) {
    return sqr(texelFetch(colortex10, UV, 0).w);
}
vec2 FetchHistoryVariance(ivec2 UV) {
    return texelFetch(colortex12, UV, 0).zw;
}
vec2 FetchHistoryAux(ivec2 UV) {
    return texelFetch(colortex13, UV, 0).xy;
}

#include "/lib/light/sh.glsl"

SphericalHarmonics FetchSH_Current(ivec2 uv) {
    return SphericalHarmonics(texelFetch(colortex6, uv, 0), texelFetch(colortex7, uv, 0).xy);
}
SphericalHarmonics FetchSH_History(ivec2 uv) {
    return SphericalHarmonics(texelFetch(colortex11, uv, 0), texelFetch(colortex12, uv, 0).xy);
}

float spatialEstimate(inout SphericalHarmonics color, inout vec2 variance, vec2 uv, vec4 GBuffer) {
    float sumWeight     = 1.0;
    float maxLum        = getLuma(ToIrradiance(color));

    ivec2 pixelUV       = ivec2(uv * viewSize);

    const int r = 3;

    for (int x = -r; x <= r; x++) {
        for (int y = -r; y <= r; y++) {
            if (x == 0 && y == 0) continue;

            ivec2 UV    = pixelUV + ivec2(x, y);
            if (!UV_WithinBounds(UV)) continue;
            vec4 GB     = FetchUnpackedGbuffer(UV);

            float depthDelta = distance(GB.w, GBuffer.w) * far;

            if (depthDelta < 2.0) {
                float weight    = pow(max0(dot(GBuffer.xyz, GB.xyz)), 2.0);

                SphericalHarmonics currentColor = FetchSH_Current(UV);
                float currentLuma   = getLuma(ToIrradiance(currentColor));

                maxLum          = max(maxLum, currentLuma);
                accumulateSH(color, currentColor, weight);
                variance       += vec2(currentLuma * weight, sqr(currentLuma) * sqr(weight));
                sumWeight      += weight;
            }
        }
    }

    divideSH(color, sumWeight);
    variance    /= vec2(sumWeight, sqr(sumWeight));

    return clamp(maxLum, 0.0, 2.71) * rpi;
}

const float gaussKernel[2][2] = {
	{ 1.0 / 4.0, 1.0 / 8.0  },
	{ 1.0 / 8.0, 1.0 / 16.0 }
};

#include "/lib/offset/gauss.glsl"

float getIrradianceLuma(vec4 shLuma) {
    return shLuma.w / 0.282095;
}

vec2 computeVariance(ivec2 pos) {
    float sumMsqr   = 0.0;
    float sumMean   = 0.0;

    for (int i = 0; i<9; i++) {
        ivec2 deltaPos     = kernelO_3x3[i];

        float lum   = getIrradianceLuma(texelFetch(colortex6, pos + deltaPos, 0));

        sumMsqr    += sqr(lum);
        sumMean    += lum;
    }
    sumMsqr  /= 9.0;
    sumMean  /= 9.0;

    return vec2(abs(sumMsqr - sqr(sumMean)) * rcp(max(sumMean, 1e-20)), sumMean);
}

SphericalHarmonics spatialColor(vec2 uv, vec4 GBuffer) {
    ivec2 pixelUV       = ivec2(uv * viewSize);
    SphericalHarmonics color = FetchSH_Current(pixelUV);

    float lumaCenter    = getLuma(ToIrradiance(color));

    #ifndef SVGF_PREACCUM
    return color;
    #endif

    vec2 variance       = computeVariance(pixelUV);

    float sigmaL        = 1.0 / (variance.x + 0.16);

    float sumWeight     = 1.0 /* gaussKernel[0][0]*/;
        scaleSH(color, sumWeight);

	const int r = 2;
	for(int y = -r; y <= r; ++y) {
		for(int x = -r; x <= r; ++x) {
            if (x == 0 && y == 0) continue;

            ivec2 UV    = pixelUV + ivec2(x, y);
            if (!UV_WithinBounds(UV)) continue;

            vec4 GB     = FetchUnpackedGbuffer(UV);

            float depthDelta = distance(GB.w, GBuffer.w) * far;

            if (depthDelta < 2.0) {
                SphericalHarmonics currentColor = FetchSH_Current(UV);

                float lum       = getLuma(ToIrradiance(currentColor));

                float distLum   = abs(lumaCenter - lum);
                    distLum     = sqr(distLum) / max(variance.y, 0.027);
                    distLum     = clamp(distLum, 0.0, pi);

                float weight    = pow(max0(dot(GBuffer.xyz, GB.xyz)), 8.0);
                    weight     *= exp(-depthDelta - sqrt(distLum * sigmaL));
                    //weight     *= gaussKernel[abs(x)][abs(y)];

                accumulateSH(color, currentColor, weight);
                sumWeight      += weight;
            }
        }
    }

    divideSH(color, sumWeight);

    return color;
}

vec3 clampNormal(vec3 normal, vec3 view){
    float NoV = saturate(dot(normal, -view));
    return normalize(NoV * view + normal);
}

void main() {
    ivec2 UV    = ivec2(coord * viewSize);
    currentSH   = vec4(0);
    currentCC   = vec4(0);
    GbufferHistory  = vec4(0, 0, 1, 1);
    historySH   = vec4(0);
    historyCC   = vec4(0);
    historyAux  = vec4(0, 0, texelFetch(colortex13, UV, 0).zw);

    vec2 lowresCoord    = coord / indirectResScale;
    ivec2 UV_LR         = ivec2(floor(coord * viewSize) / indirectResScale);
    float sceneDepth    = texelFetch(depthtex0, UV_LR, 0).x;

    if (landMask(sceneDepth) && UV_WithinBounds(UV)) {
        vec2 coord      = saturate(lowresCoord);
        vec2 scaledCoord = coord * indirectResScale;

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth), false);
        vec3 scenePos   = viewToSceneSpace(viewPos);

        bool hand       = sceneDepth < texelFetch(depthtex2, UV_LR, 0).x;

        vec3 reprojection   = reproject(scenePos, false);
        bool offscreen      = saturate(reprojection) != reprojection;

        vec2 scaledReprojection = reprojection.xy * indirectResScale;

        // Read Current Frame Geometry Data
        vec4 currentGBuffer     = FetchUnpackedGbuffer(UV);
        float currentDistance   = saturate(length(scenePos) / far);

        // Store Current Frame Geometry Data
        GbufferHistory.xyz  = FetchUnpackedGbufferRaw(UV).xyz;
        GbufferHistory.w    = sqrt(currentDistance);

        // Depth Rejection
        vec3 cameraMovement = mat3(gbufferModelView) * (cameraPosition - previousCameraPosition);

        bool doSpatialEstimate = false;

        float currentAdapt  = texelFetch(colortex7, UV, 0).z;

        if (offscreen) {
            doSpatialEstimate = true;
        } else {
            SphericalHarmonics previousColor = InitSH();

            vec2 PrevVariance   = vec2(0);
            vec2 PrevAux        = vec2(0);

            // Sample History
            ivec2 repPixel  = ivec2(floor(scaledReprojection * viewSize - vec2(0.5)));
            vec2 subpix     = fract(scaledReprojection * viewSize - vec2(0.5) - repPixel);

            const ivec2 offset[4] = ivec2[4](
                ivec2(0, 0),
                ivec2(1, 0),
                ivec2(0, 1),
                ivec2(1, 1)
            );

            float weight[4]     = float[4](
                (1.0 - subpix.x) * (1.0 - subpix.y),
                subpix.x         * (1.0 - subpix.y),
                (1.0 - subpix.x) * subpix.y,
                subpix.x         * subpix.y
            );

            float sumWeight     = 0.0;

            for (uint i = 0; i<4; i++) {
                ivec2 UV            = repPixel + offset[i];
                if (!UV_WithinBounds(UV)) continue;

                float depthDelta    = distance(FetchHistoryDistance(UV), currentDistance) - abs(cameraMovement.z / far);
                bool depthRejection = depthDelta / abs(currentDistance) < 0.1;

                if (depthRejection) {
                    accumulateSH(previousColor, FetchSH_History(UV), weight[i]);
                    PrevVariance   += FetchHistoryVariance(UV) * weight[i];
                    PrevAux        += FetchHistoryAux(UV) * weight[i];
                    sumWeight      += weight[i];
                }
            }

            if (sumWeight > 1e-3) {
                divideSH(previousColor, sumWeight);
                PrevVariance       /= sumWeight;
                PrevAux            /= sumWeight;

                float frames        = min(PrevAux.x * maxFrames + 1.0, maxFrames);
                float alphaColor    = max(0.025 * minAccumMult, 1.0 / frames);
                float alphaVariance = max(0.05 * minAccumMult, 1.0 / frames);

                float adaptDelta    = sqr(max0(distance(currentAdapt, PrevAux.y) - 1.0 / 1024.0) / avgOf(vec2(currentAdapt, PrevAux.y) + 5e-3));

                float rejection     = 1.0 / (1.0 + adaptDelta);
                    rejection       = saturate(1.0 - rejection);

                    alphaColor      = max(saturate(0.66 * rejection * ADAPT_STRENGTH), alphaColor);
                    alphaVariance   = max(saturate(0.75 * rejection * ADAPT_STRENGTH), alphaVariance);

                SphericalHarmonics Color = spatialColor(scaledCoord, currentGBuffer);

                float currentLuma   = getLuma(ToIrradiance(Color));

                vec2 currentVar     = vec2(currentLuma, sqr(currentLuma));
                vec2 Variance       = mix(PrevVariance, currentVar, alphaVariance);

                Color               = mixSH(previousColor, Color, alphaColor);

                currentSH           = Color.Y;
                currentCC           = vec4(Color.CoCg, sqrt(max0(Variance.y - sqr(Variance.x))) / max(Variance.x, 1e-8), frames / maxFrames);
                currentCC.z        *= max(1.0 + rejection, 4.0 / frames);   //Boost Variance on new Pixels

                historySH           = Color.Y;
                historyCC.xy        = Color.CoCg;
                historyCC.zw        = Variance;

                historyAux.x        = frames / maxFrames;
                historyAux.y        = mix(PrevAux.y, currentAdapt, max(0.2, 1.0 / frames));
            } else {
                doSpatialEstimate = true;
            }
        }

        if (doSpatialEstimate) {
            SphericalHarmonics Color = FetchSH_Current(UV);

            currentCC.z     = spatialEstimate(Color, historyCC.zw, scaledCoord, currentGBuffer);
            
            currentSH       = Color.Y;
            currentCC.xy    = Color.CoCg;
            
            historySH       = Color.Y;
            historyCC.xy    = Color.CoCg;

            historyAux.y    = currentAdapt;
        }
    }

    currentSH   = clamp16FN(currentSH);
    currentCC   = clamp16FN(currentCC);
    historySH   = clamp16FN(historySH);
    historyCC   = clamp16FN(historyCC);
}