#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/colorspace.glsl"
#include "/lib/shadowconst.glsl"

layout(location = 0) out vec4 data0;
layout(location = 1) out vec4 data1;

#define gSHADOW

uniform sampler2D tex;

in voxelGeometryOut {
    flat int isVoxel;
    flat uvec4 voxelWrite;
} voxelIn;

in shadowGeometryOut {
    flat int matID;
    float warp;
    vec2 uv;
    vec3 tint;
    vec3 scenePos;
    vec3 worldPos;
} shadowIn;


/* ------ caustics ------ */

#define FUTIL_MAT16
#include "/lib/fUtil.glsl"

uniform sampler2D noisetex;

uniform float frameTimeCounter;

uniform vec3 lightDir;

#include "/lib/util/bicubic.glsl"

const float viewDist = 0.0;

#include "/lib/atmos/waterConst.glsl"

#include "/lib/atmos/waterWaves.glsl"

vec3 waterNormal(vec3 position, float dstep, int matID) {
        dstep       = max(dstep, 0.015);

    vec2 delta;
        delta.x     = waterWaves(position + vec3( dstep, 0.0, -dstep), matID);
        delta.y     = waterWaves(position + vec3(-dstep, 0.0,  dstep), matID);
        delta      -= waterWaves(position + vec3(-dstep, 0.0, -dstep), matID);

    return normalize(vec3(-delta.x, 2.0 * dstep, -delta.y));
}

float projectedCaustic(vec3 pos, vec3 normal, vec3 lightDir) {
    vec3 dPdx   = dFdx(pos);
    vec3 dPdy   = dFdy(pos);

    float num   = dotSelf(dPdx) * dotSelf(dPdy);

    vec3 refractLight = refract(-lightDir, normal, rcp(1.33));
        dPdx   += 2.0 * dFdx(refractLight);
        dPdy   += 2.0 * dFdy(refractLight);

    float denom = dotSelf(dPdx) * dotSelf(dPdy);

    return sqrt(num * rcp(denom));
}

void main() {
    if (shadowIn.matID == 31) discard;

    if (voxelIn.isVoxel == 1) {

        data0   = vec4(voxelIn.voxelWrite) / 65535.0;
        data1   = vec4(1.0, 0.0, 0.0, 0.0);

    } else {

        #ifdef generateShadowmap
        if (shadowIn.matID == 102 || shadowIn.matID == 103) { 
            vec3 waves      = waterNormal(shadowIn.worldPos, rcp(shadowMapResolution) * shadowIn.warp * 2.0, shadowIn.matID);
            vec3 refractLight = refract(lightDir, waves, rcp(1.33));
            float caustic   = projectedCaustic(shadowIn.scenePos, waves, lightDir);
            data0           = vec4(caustic * 0.25, 1.0, 1.0, 1.0);
        } else {
            data0           = texture(tex, shadowIn.uv, -shadowmapTextureMipBias);
            if (data0.a<0.1) discard;
                data0.rgb  *= shadowIn.tint;

            data0.rgb       = linearToAP1Albedo(toLinear(data0.rgb));
        }

        data1   = vec4(float(shadowIn.matID == 102 || shadowIn.matID == 103), 0.0, 0.0, 0.0);

        #else

        data0   = vec4(0);
        data1   = vec4(0);

        #endif
    }
}