#version 450 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 0,3,6,7,8 */
layout(location = 0) out vec3 sceneColor;
layout(location = 1) out vec4 reflectionColor;
//layout(location = 2) out vec4 GBuffer;
layout(location = 2) out vec4 fresnelColor;
layout(location = 3) out vec3 fogScattering;
layout(location = 4) out vec3 fogTransmittance;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat2x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex7;
uniform sampler2D colortex8;
uniform sampler2D colortex15;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform int frameCounter;
uniform int isEyeInWater;
uniform int worldTime;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float lightFlip;
uniform float sunAngle;
uniform float rainStrength, wetness;
uniform float worldAnimTime;

uniform ivec2 eyeBrightness;
uniform ivec2 eyeBrightnessSmooth;

uniform vec2 taaOffset;
uniform vec2 viewSize, pixelSize;

uniform vec3 cameraPosition;
uniform vec3 lightDir, lightDirView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

/* ------ INCLUDES ------ */
#define FUTIL_MAT16
#define FUTIL_TBLEND
#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/atmos/air/const.glsl"
#include "/lib/atmos/phase.glsl"
#include "/lib/atmos/waterConst.glsl"
#include "/lib/frag/noise.glsl"

#include "/lib/light/cacheLib.glsl"


/* ------ VOLUMETRIC FOG ------ */

#ifdef freezeAtmosAnim
    const float fogTime   = float(atmosAnimOffset) * 0.006;
#else
    float fogTime     = frameTimeCounter * 0.6;
#endif

vec3 smokeDensity(vec3 rPos, float altitude) {
    #ifdef netherSmokeEnabled
    float smokeFade = sqr(1.0 - linStep(altitude, 128.0, 256.0));
        smokeFade  *= expf(-max0((altitude - 32.0) * 0.015)) * 0.6 + 0.4;

    float smoke     = 0.0;
    float smokeGlowing = 0.0;

    if (smokeFade > 1e-10) {
        vec3 wind   = fogTime * vec3(1.0, -0.9, 0.1);

        vec3 pos    = (rPos + cameraPosition) * 0.03;
            pos    += value3D(pos * 4.0 + wind * 0.5) * 1.5 - 0.75;
            pos    += value3D(pos * 8.0 - wind) - 0.5;

        pos.x *= 0.7;

        float noise     = value3D(pos * 4.0 + wind);
            noise      += value3D(pos * 8.0 + wind * 2 + noise * 0.5) * 0.5;
            //noise      += value3D(pos * 16.0 + wind * 4 + noise * 0.5) * 0.25;
            noise      /= 1.5;

        smoke       = smokeFade - noise;
        float glowProximity = sqrt(sstep(length(rPos), 4.0, 24.0));
        smokeGlowing = saturate((glowProximity * (0.3 + cube(smokeFade) * 0.15) - noise) * pi);
        //smoke      -= 0.1;

        smoke       = cubeSmooth(max0(smoke)) * (smokeFade * 0.5 + 0.5);
        smokeGlowing *= smokeFade;
    }

    return vec3(1.0, smoke, smokeGlowing);
    #else
    return vec3(1.0, 0.0, 0.0);
    #endif
}

vec3 readVolumeLight(ivec3 index) {
    if (outsideCacheVolume(index)) return vec3(0);
    return clamp16F(texelFetch(colortex15, getCachePixelCacheSpace(index, 6), 0).rgb);
}
vec3 readVolumeLight(ivec3 index, vec3 skylight) {
    if (outsideCacheVolume(index)) return skylight;
    return clamp16F(texelFetch(colortex15, getCachePixelCacheSpace(index, 6), 0).rgb);
}

vec3 readVolumeLightSmooth(vec3 scenePos, vec3 skylight) {
    ivec3 voxelIndex = sceneToCacheSpace(scenePos);

    #ifdef radianceVolumeSmoothRead
        vec3 alpha  = fract(scenePos + cameraPosition - 0.5);
            alpha   = linStep(alpha, 0.001, 0.999);

        vec3 direction = fract(scenePos + cameraPosition) - 0.5;
        bvec3 isPositive = greaterThan(direction, vec3(0));


        /*
            |    | Y+ |    |

            | X- | M  | X+ |

            |    | Y- |    |
        */

        ivec3 mid   = voxelIndex;
        ivec3 edge  = voxelIndex + ivec3(isPositive.x ? 1 : -1, isPositive.y ? 1 : -1, isPositive.z ? 1 : -1);

        alpha.x         = isPositive.x ? alpha.x : 1.0 - alpha.x;
        alpha.y         = isPositive.y ? alpha.y : 1.0 - alpha.y;
        alpha.z         = isPositive.z ? alpha.z : 1.0 - alpha.z;

        vec3 tap_LY_LZ  = mix(readVolumeLight(ivec3(mid.x, mid.y, mid.z), skylight), 
                            readVolumeLight(ivec3(edge.x, mid.y, mid.z), skylight), alpha.x);
        vec3 tap_LY_HZ  = mix(readVolumeLight(ivec3(mid.x, mid.y, edge.z), skylight), 
                            readVolumeLight(ivec3(edge.x, mid.y, edge.z), skylight), alpha.x);

        vec3 tap_HY_LZ  = mix(readVolumeLight(ivec3(mid.x, edge.y, mid.z), skylight), 
                            readVolumeLight(ivec3(edge.x, edge.y, mid.z), skylight), alpha.x);
        vec3 tap_HY_HZ  = mix(readVolumeLight(ivec3(mid.x, edge.y, edge.z), skylight), 
                            readVolumeLight(ivec3(edge.x, edge.y, edge.z), skylight), alpha.x);

        vec3 tap_LZ     = mix(tap_LY_LZ, tap_HY_LZ, alpha.y);
        vec3 tap_HZ     = mix(tap_LY_HZ, tap_HY_HZ, alpha.y);

        return mix(tap_LZ, tap_HZ, alpha.z);
    #else
        return readVolumeLight(voxelIndex, skylight);
    #endif
}

flat in vec3 fogCol;
flat in vec3 fogCol2;
flat in vec3 fogEmissionColor;

const vec3 hazeCoeff    = vec3(6e-3) * netherHazeDensity;
const vec3 smokeCoeff   = vec3(3e-2) * netherSmokeDensity;

mat2x3 volumetricFog(vec3 scenePos, vec3 sceneDir, float dither) {
    vec3 startPos   = gbufferModelViewInverse[3].xyz;
    vec3 endPos     = scenePos;
        if (length(endPos) > 192.0) endPos = sceneDir * 192.0;

    float baseStep  = length(endPos - startPos);
    float stepCoeff = saturate(baseStep * rcp(clamp(far, 128.0, 192.0)));

    uint steps      = 8 + uint(stepCoeff * 16.0);

    vec3 rStep      = (endPos - startPos) / float(steps);
    vec3 rPos       = startPos + rStep * dither;
    float rLength   = length(rStep);

    mat3x3 scattering = mat3x3(0.0);
    vec3 transmittance = vec3(1.0);

    const mat2x3 scatterMat   = mat2x3(hazeCoeff, smokeCoeff);
    const mat2x3 extinctMat   = mat2x3(hazeCoeff * 1.1, smokeCoeff);

    for (uint i = 0; i < steps; ++i, rPos += rStep) {
        if (maxOf(transmittance) < 0.0005) break;

        if (length(rPos) > length(endPos)) continue;

        float altitude  = rPos.y + eyeAltitude;

        if (altitude > 256.0) continue;

        vec3 density    = smokeDensity(rPos, altitude);

        vec2 stepRho    = density.xy * rLength;
        vec3 od         = extinctMat * stepRho;

        vec3 stepT      = expf(-od);
        vec3 scatterInt = saturate((stepT - 1.0) * rcp(-max(od, 1e-16)));
        vec3 visScatter = transmittance * scatterInt;

        vec3 a          = visScatter * transmittance;

        mat2x3 stepScatter = mat2x3(scatterMat[0] * stepRho.x * a, scatterMat[1] * stepRho.y * a);

        vec3 lightScatter   = readVolumeLightSmooth(rPos, vec3(0.0));

        #ifdef netherSmokeEmission
            vec3 smokeEmission = mix(vec3(1.0, 0.1, 0.0), fogEmissionColor * 0.5, saturate(density.z));
            smokeEmission   = smokeEmission * stepRho.y * a * density.z;

            scattering     += mat3x3(stepScatter[0] * (fogCol2 + lightScatter), stepScatter[1] * (fogCol + lightScatter), smokeEmission);
        #else
            scattering     += mat3x3(stepScatter[0] * (fogCol2 + lightScatter), stepScatter[1] * (fogCol + lightScatter), vec3(0.0));
        #endif

        transmittance  *= stepT;
    }

    //vec3 smokeCol       = mix(fogCol, lightColor[1], saturate(scattering[1]));

    vec3 color          = scattering[0] + scattering[1] + scattering[2] * 0.6 * netherSmokeEmissionMult;


    if (color != color) {   //because NaNs on nVidia don't need a logic cause to happen
        color = vec3(0.0);
        transmittance = vec3(1.0);
    }

    return mat2x3(color, saturate(transmittance));
}

void applyFogData(inout vec3 color, in mat2x3 data) {
    color = color * data[1] + data[0];
}

#include "/lib/atmos/fog.glsl"

/* ------ REFRACTION ------ */
vec3 refract2(vec3 I, vec3 N, vec3 NF, float eta) {     //from spectrum by zombye
    float NoI = dot(N, I);
    float k = 1.0 - eta * eta * (1.0 - NoI * NoI);
    if (k < 0.0) {
        return vec3(0.0); // Total Internal Reflection
    } else {
        float sqrtk = sqrt(k);
        vec3 R = (eta * dot(NF, I) + sqrtk) * NF - (eta * NoI + sqrtk) * N;
        return normalize(R * sqrt(abs(NoI)) + eta * I);
    }
}


/* ------ REFLECTIONS ------ */

#include "/lib/atmos/project.glsl"
#include "/lib/util/colorspace.glsl"
#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/material.glsl"
#include "/lib/brdf/specular.glsl"
#include "/lib/frag/capture.glsl"

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}


#define DIM -1
#include "/lib/brdf/reflections.glsl"

float ditherBluenoiseSlow() {
    ivec2 coord = ivec2(gl_FragCoord.xy);
    float noise = texelFetch(noisetex, coord & 255, 0).a;

    #ifdef taaEnabled
        //noise   = fract(noise + float(frameCounter)/32);
    #endif

    return noise;
}

vec2 ditherBluenoiseFull() {
    ivec2 coord = ivec2(gl_FragCoord.xy);
    /*
    vec2 noise = texelFetch(depthtex2, ivec3(coord & 255, 0), 0).xy;

    #ifdef taaEnabled
        noise   = fract(noise + torroidalShift.xy * float(frameCounter % 4096));
    #endif*/

    vec2 noise  = texelFetch(depthtex2, ivec3(coord & 255, frameCounter & 7), 0).xy;
        noise   = fract(noise + torroidalShift.xy * floor(frameCounter / 8));

    return noise;
}

vec3 ditherBluenoiseStaticF3() {
    ivec2 coord = ivec2(gl_FragCoord.xy);

    vec3 noise  = texelFetch(depthtex2, ivec3(coord & 255, 0), 0).xyz;

    return noise;
}

void main() {
    sceneColor  = stex(colortex0).rgb;

    ivec2 texelUV   = ivec2(coord * viewSize);

    GData00 Gbuffer0 = ReadGbuffer0(texelUV);

    vec2 sceneDepth = vec2(stex(depthtex0).x, stex(depthtex1).x);

    vec3 viewPos0   = screenToViewSpace(vec3(coord, sceneDepth.x));
    vec3 scenePos0  = viewToSceneSpace(viewPos0);

    vec3 viewPos1   = screenToViewSpace(vec3(coord, sceneDepth.y));
    vec3 scenePos1  = viewToSceneSpace(viewPos1);

    vec3 viewDir    = normalize(viewPos0);
    vec3 worldDir   = normalize(scenePos0);

    bool translucent    = sceneDepth.x < sceneDepth.y;

    #ifdef refractionEnabled
    if (translucent && Gbuffer0.BlockID.x != 100){
        vec3 viewNormal     = mat3(gbufferModelView) * Gbuffer0.SceneNormal;
        //vec3 flatNormal     = normalize(cross(dFdx(scenePos0), dFdy(scenePos0)));
        vec3 flatNormal     = decodeNormal(unpack2x8(texelFetch(colortex2, texelUV, 0).z));

        if (flatNormal != flatNormal) flatNormal = Gbuffer0.SceneNormal;
            flatNormal = clampDIR(flatNormal);

        vec3 flatViewNormal = mat3(gbufferModelView) * flatNormal;

        vec3 normalCorrected = dot(viewNormal, viewDir) > 0.0 ? -viewNormal : viewNormal;

        vec3 refractedDir   = refract2(normalize(viewPos1), normalCorrected, flatViewNormal, rcp(1.33));
        //vec3 refractedDir   = refract(normalize(viewPos1), normalCorrected, rcp(1.33));

        float refractedDist = distance(viewPos0, viewPos1);

        vec3 refractedPos   = viewPos1 + refractedDir * refractedDist;

        vec3 screenPos      = viewToScreenSpace(refractedPos);

        float distToEdge    = maxOf(abs(screenPos.xy * 2.0 - 1.0));
            distToEdge      = sqr(sstep(distToEdge, 0.7, 1.0));

            screenPos.xy    = mix(screenPos.xy, coord, distToEdge);

        //vec2 refractionDelta = coord - screenPos.xy;

        float sceneDepthNew = texture(depthtex1, screenPos.xy).x;

        if (sceneDepthNew > sceneDepth.x) {
            sceneDepth.y    = sceneDepthNew;
            viewPos1        = screenToViewSpace(vec3(screenPos.xy, sceneDepth.y));
            scenePos1       = viewToSceneSpace(viewPos1);

            sceneColor.rgb  = texture(colortex0, screenPos.xy).rgb;
        }
    }
    #endif

    float bluenoise     = ditherBluenoise();

    bool water      = Gbuffer0.BlockID.x == 102;

    mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, ivec2(coord * viewSize), 0));

    float dist0     = distance(scenePos0, gbufferModelViewInverse[3].xyz);
    float dist1     = distance(scenePos1, gbufferModelViewInverse[3].xyz);

    if (translucent && isEyeInWater == 0) {

        if (water) {
            sceneColor  = waterFog(sceneColor, distance(scenePos0, scenePos1), lightColor[1]);
        }

        sceneColor.rgb = blendTranslucencies(sceneColor.rgb, stex(colortex3), lightingData[1]);
    }

    sceneColor      = clamp16F(sceneColor);

    fogScattering   = vec3(0.0);
    fogTransmittance = vec3(1.0);

    #ifdef fogVolumeEnabled
        const float fogLOD      = 2.0;

        vec2 fogCoord           = coord * FOG_LOWRES_FACTOR;

        if (clamp(fogCoord, -pixelSize * FOG_LOWRES_FACTOR, 1.0 + pixelSize * FOG_LOWRES_FACTOR) == fogCoord) {
            fogCoord         = saturate(fogCoord);

            float sceneDepth = texelFetch(depthtex0, ivec2(fogCoord * viewSize), 0).x;

            vec3 viewPos0   = screenToViewSpace(vec3(fogCoord, sceneDepth));
            vec3 scenePos0  = viewToSceneSpace(viewPos0);

            vec3 viewDir    = normalize(viewPos0);
            vec3 worldDir   = normalize(scenePos0);

            float vDotL     = dot(viewDir, lightDirView);
            bool isSky      = !landMask(sceneDepth);

            mat2x3 fogData  = mat2x3(vec3(0.0), vec3(1.0));

            if (isEyeInWater == 0) {
                fogData    = volumetricFog(scenePos0, worldDir, bluenoise);
            }

            fogScattering = clamp16F(fogData[0]);
            fogTransmittance = saturate(fogData[1]);
        }
    #endif

    /*
        Reflections
    */

    reflectionColor     = vec4(0);
    fresnelColor        = vec4(0);
    //GBuffer             = vec4(0,0,1,1);

    vec2 uvLR           = coord / reflectionResScale;
    ivec2 pixelPos      = ivec2(floor(coord * viewSize) / reflectionResScale);
    float sceneDepthLR  = texelFetch(depthtex0, pixelPos, 0).x;

    if (landMask(sceneDepthLR.x)) {
        ivec2 UV        = ivec2(uvLR * viewSize);

        vec3 viewPos0   = screenToViewSpace(vec3(uvLR, sceneDepthLR));
        vec3 scenePos0  = viewToSceneSpace(viewPos0);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        //GBuffer         = saturate(vec4(Gbuffer0.SceneNormal * 0.5 + 0.5, sqrt(depthLinear(sceneDepthLR))));

        vec3 viewNormal = mat3(gbufferModelView) * Gbuffer0.SceneNormal;

        bool water      = Gbuffer0.BlockID.x == 102;

        mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, UV, 0));

        vec3 viewDir    = normalize(viewPos0);

        mat2x3 reflectionAux = lightingData;

        materialProperties material = decodeLabBasic(Gbuffer1.MaterialData.xy);
        if (water) material = materialProperties(0.0001, 0.02, false, false, mat2x3(0.0));

        if (dot(viewDir, viewNormal) > 0.0) viewNormal = -viewNormal;

        vec3 reflectDir = reflect(viewDir, viewNormal);
        
        vec4 reflection = vec4(0.0);
        vec3 fresnel    = vec3(0.0);

        vec3 blueNoiseF3 = ditherBluenoiseStaticF3();

        float skyOcclusion  = cubeSmooth(sqr(linStep(Gbuffer0.Lightmap.y, skyOcclusionThreshold - 0.2, skyOcclusionThreshold)));

        #ifdef resourcepackReflectionsEnabled
            /* --- ROUGH REFLECTIONS --- */

            float roughnessFalloff  = 1.0 - (linStep(material.roughness, roughnessThreshold * 0.71, roughnessThreshold));

            #ifdef roughReflectionsEnabled
            if (material.roughness < 0.001 || water) {
            #endif

                vec3 reflectSceneDir = mat3(gbufferModelViewInverse) * reflectDir;

                /*
                bool hit    = false;

                #ifdef screenspaceReflectionsEnabled
                    vec3 reflectedPos   = screenspaceRT(viewPos0, reflectDir, ditherBluenoise());
                        hit    = reflectedPos.z < 1.0;

                    if (hit) reflection += vec4(texelFetch(colortex0, ivec2(reflectedPos.xy * viewSize), 0).rgb, 1.0);
                #endif

                #ifdef voxelPT_ReflectionsEnabled
                    if (!hit) {
                        vec4 voxelSample = reflectionTrace(viewPos0, viewNormal, reflectDir);
                        hit = voxelSample.a > 0.5;
                        if (hit) reflection += vec4(voxelSample.rgb, 1.0);
                    }
                #endif
                */

                vec4 voxelSample = reflectionMultibounce(viewPos0, viewNormal, reflectDir, blueNoiseF3, material, skyOcclusion);
                reflection += voxelSample;

                if (clamp16F(reflection) != reflection) reflection = vec4(0.0);

                fresnel    += BRDFfresnelAlbedoTint(-viewDir, viewNormal, material, reflectionAux[1]);
                //fresnel     = getRoughFresnel(viewNormal, viewDir, blueNoiseF3.xy, material, reflectionAux[1]);

            #ifdef roughReflectionsEnabled
            
            } else {
                mat3 rot        = getRotationMat(vec3(0, 0, 1), viewNormal);
                vec3 tangentV   = viewDir * rot;
                //float noise     = ditherBluenoiseSlow();
                float dither    = ditherGradNoiseTemporal();

                const uint steps    = roughReflectionSamples;
                const float rSteps  = 1.0 / float(steps);

                for (uint i = 0; i < steps; ++i) {
                    if (roughnessFalloff <= 1e-3) break;
                    vec3 roughNrm   = rot * ggxFacetDist(-tangentV, (material.roughness), blueNoiseF3.xy);

                    vec3 reflectDir = reflect(viewDir, roughNrm);
                    blueNoiseF3     = fract(blueNoiseF3 + torroidalShift.xyz);

                    //vec3 reflectSceneDir = mat3(gbufferModelViewInverse) * reflectDir;

                    /*
                    bool hit    = false;

                    #ifdef screenspaceReflectionsEnabled
                        vec3 reflectedPos       = vec3(1.1);
                        if (material.roughness < 0.3) reflectedPos = screenspaceRT(viewPos0, reflectDir, blueNoiseF3.z);
                        else if (material.roughness < 0.7) reflectedPos = screenspaceRT_LR(viewPos0, reflectDir, blueNoiseF3.z);

                        hit    = reflectedPos.z < 1.0;

                        if (hit) reflection += vec4(texelFetch(colortex0, ivec2(reflectedPos.xy * viewSize), 0).rgb, 1.0);
                    #endif

                    #ifdef voxelPT_ReflectionsEnabled
                        if (!hit) {
                            vec4 voxelSample = reflectionTrace(viewPos0, roughNrm, reflectDir);
                            hit = voxelSample.a > 0.5;
                            if (hit) reflection += vec4(voxelSample.rgb, 1.0);
                        }
                    #endif
                    */


                    vec4 voxelSample = reflectionMultibounce(viewPos0, roughNrm, reflectDir, blueNoiseF3, material, skyOcclusion);
                    reflection += voxelSample;

                    fresnel    += BRDFfresnelAlbedoTint(-viewDir, roughNrm, material, reflectionAux[1]);
                }
                if (clamp16F(reflection) != reflection) reflection = vec4(0.0);

                reflection *= rSteps;
                fresnel    *= rSteps;

                reflection.a *= roughnessFalloff;
            }

            #else
                reflection.a *= roughnessFalloff;
            #endif
        
        #else
            /* --- WATER REFLECTIONS --- */
            if (water) {
                vec3 reflectSceneDir = mat3(gbufferModelViewInverse) * reflectDir;

                vec3 reflectedPos = screenspaceRT(viewPos0, reflectDir, ditherBluenoise());
                if (reflectedPos.z < 1.0) reflection += vec4(texelFetch(colortex0, ivec2(reflectedPos.xy * viewSize), 0).rgb, 1.0);

                if (clamp16F(reflection) != reflection) reflection = vec4(0.0);

                    fresnel    += BRDFfresnel(-viewDir, viewNormal, material, reflectionAux[1]);
            }
        #endif

        if (material.conductorComplex && material.conductor) fresnel = (fresnel / sqrt(max(reflectionAux[1], 1e-8))) * rpi;

        reflectionColor     = clamp16F(reflection);
        fresnelColor        = clamp16F(vec4(fresnel, material.roughness));
    }
}