#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 0,3 */
layout(location = 0) out vec3 sceneColor;
layout(location = 1) out vec4 bloomTiles;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat2x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex8;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;

uniform int frameCounter;
uniform int isEyeInWater;

uniform float near, far;
uniform float lightFlip;
uniform float sunAngle;

uniform vec2 taaOffset;
uniform vec2 pixelSize, viewSize;
uniform vec2 skyCaptureResolution;

uniform vec3 lightDir, lightDirView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

void applyFogData(inout vec3 color, in mat2x3 data) {
    color = color * data[1] + data[0];
}

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/util/transforms.glsl"


#include "/lib/atmos/air/const.glsl"
#include "/lib/atmos/waterConst.glsl"

#include "/lib/atmos/fog.glsl"


/* REFLECTIONS */
#include "/lib/util/colorspace.glsl"
#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/material.glsl"
#include "/lib/brdf/specular.glsl"
#include "/lib/frag/capture.glsl"


/*
    Based on the bilateral blur chocapic13 does.
    Actually seems illegally more effective than my checkerboard solution. Can't make this shit up...
*/

#define FOG_SCATTER_IMAGE colortex7
#define FOG_TRANSMITTANCE_IMAGE colortex8

void BilateralFog(out vec3 scattering, out vec3 transmittance) {
    scattering  = vec3(0);
    transmittance = vec3(0);

    vec3 zmult      = vec3(1.0/(far * near),far+near,far-near);

    ivec2 depthUV   = ivec2(coord * viewSize);
    ivec2 fogUV     = ivec2(coord * viewSize / FOG_LOWRES_FACTOR);
    ivec2 jitter    = (ivec2(gl_FragCoord.xy + frameCounter) % 2) * 2;

    float depthCenter = depthLinear(texelFetch(depthtex0, depthUV, 0).x);

    float sumWeight = 0.0;

    ivec2 depthTapUV = depthUV + ivec2(-2) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
    float depthTap  = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
    float weight    = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
    ivec2 fogTapUV  = fogUV + ivec2(-2) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(-2, 0) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = fogUV + ivec2(-2, 0) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(0) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = fogUV + ivec2(0) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(0, -2) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = fogUV + ivec2(0, -2) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

    scattering     /= sumWeight;
    transmittance  /= sumWeight;

    return;
}

#define colorSampler colortex3
#define fresnelSampler colortex6

vec4 FetchGbuffer(ivec2 UV) {
        UV          = ivec2(vec2(UV) / reflectionResScale);
    vec3 normals    = decodeNormal(texelFetch(colortex1, UV, 0).xy);
    float depth     = depthLinear(texelFetch(depthtex0, UV, 0).x) * far;
    return vec4(normals, depth);
}

float sphericalGaussian(vec3 dir1, vec3 dir2, float lambda) {
    float w = exp(lambda * (saturate(dot(dir1, dir2)) - 1.0));
    return w > 1e-8 ? w : exp(-saturate(dot(dir1, dir2)) * 16.0);
}

float lobeSharpness(float roughness) {
    return 2.0 / sqr(max(roughness, 1e-5));
}

float specularLobeWeight(vec3 centerNorm, vec3 norm, float centerRoughness, float roughness, float beta) {
    float lambaCenter = lobeSharpness(centerRoughness);
    float lambaSample = lobeSharpness(roughness);
    return pow(2.0 * sqrt(lambaCenter*lambaSample) / max(lambaCenter + lambaSample, 1e-5), beta) 
        * sphericalGaussian(centerNorm, norm, beta * lambaCenter * lambaSample / max(lambaCenter + lambaSample, 1e-5));
}

const float lobeFactor = SPF_LOBE;
const float sharpenWeight = SPF_SHARPEN;

vec4 modifySpecularColor(vec4 color) {
    return vec4(pow(length(color.rgb), sharpenWeight) * normalizeSafe(color.rgb), color.a);
}
vec4 inverseModifySpecularColor(vec4 color) {
    return vec4(pow(length(color.rgb), 1.0 / sharpenWeight) * normalizeSafe(color.rgb), color.a);
}

uniform sampler3D depthtex2;
vec3 ditherBluenoiseF3() {
    ivec2 coord = ivec2(gl_FragCoord.xy);

    vec3 noise  = texelFetch(depthtex2, ivec3(coord & 255, 0), 0).xyz;
        noise   = fract(noise + torroidalShift.xyz * frameCounter);

    return noise;
}

void specularSpatialUpscale(out vec4 reflection, out vec3 fresnel, vec2 uv, vec4 centerData, float centerRoughness) {
    reflection  = vec4(0);
    fresnel     = vec3(0);

    ivec2 UV    = ivec2(uv * viewSize * reflectionResScale);

    float totalWeight = 0.0;

    const float normalExp = 256.0;

    ivec2 jitter        = ivec2((ditherBluenoiseF3().xy - 0.5));

    const int r = 2;
	for(int y = -r; y <= r; ++y) {
		for(int x = -r; x <= r; ++x) {
			ivec2 p = UV + ivec2(x, y) + jitter;

            bool valid          = all(greaterThanEqual(p, ivec2(0))) && all(lessThan(p, ivec2(viewSize * reflectionResScale)));

            if (!valid) continue;

            vec4 currentData    = FetchGbuffer(p);

            vec4 currentColor   = texelFetch(colorSampler, p, 0);
            vec4 currentFresnel = texelFetch(fresnelSampler, p, 0);

            float w         = float(valid);
                //w          *= specularKernel[abs(x)][abs(y)];

            float distDepth = abs(centerData.a - currentData.a) * 4.0;

                //w *= pow(max(0.0, dot(centerData.rgb, currentData.rgb)), normalExp);
                w *= exp(-distDepth);
                w *= max0(specularLobeWeight(centerData.xyz, currentData.xyz, centerRoughness, currentFresnel.a, lobeFactor));

            //accumulate stuff
            reflection  += modifySpecularColor(currentColor) * w;
            fresnel     += currentFresnel.rgb * w;

            totalWeight += w;
        }
    }
    totalWeight     = max(totalWeight, 1e-16);

    reflection  /= totalWeight;
    reflection   = inverseModifySpecularColor(reflection);
    fresnel.rgb /= totalWeight;
}


void main() {
    sceneColor  = stex(colortex0).rgb;

    float sceneDepth = stex(depthtex0).x;

    vec3 viewPos0    = screenToViewSpace(vec3(coord, sceneDepth));
    vec3 viewDir    = normalize(viewPos0);
    vec3 scenePos   = viewToSceneSpace(viewPos0);

    if (landMask(sceneDepth)) {
        ivec2 UV        = ivec2(coord * viewSize);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        vec3 viewNormal = mat3(gbufferModelView) * Gbuffer0.SceneNormal;

        bool water      = Gbuffer0.BlockID.x == 102;

        mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, ivec2(coord * viewSize), 0));

        vec3 viewDir    = normalize(viewPos0);

        mat2x3 reflectionAux = lightingData;

        materialProperties material = decodeLabBasic(Gbuffer1.MaterialData.xy);
        if (water) material = materialProperties(0.0001, 0.02, false, false, mat2x3(0.0));

        if (dot(viewDir, viewNormal) > 0.0) viewNormal = -viewNormal;

        vec3 reflectDir = reflect(viewDir, viewNormal);
        
        vec4 reflection = texture(colortex3, coord * reflectionResScale);
        vec3 fresnel    = texture(colortex6, coord * reflectionResScale).rgb;

        specularSpatialUpscale(reflection, fresnel, coord, vec4(Gbuffer0.SceneNormal, depthLinear(sceneDepth) * far), material.roughness);

        if (material.conductorComplex && material.conductor) fresnel *= sqrt(reflectionAux[1]) * pi;

        #ifdef resourcepackReflectionsEnabled
            if (material.conductor) sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb * fresnel, reflection.a);
            else sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb, fresnel * reflection.a);            
        #else
            /* --- WATER REFLECTIONS --- */
            if (water) {
                sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb, fresnel * reflection.a);
            }
        #endif

        #ifndef fogVolumeEnabled
        if (isEyeInWater == 0) {
            sceneColor  = simpleFog(sceneColor, length(scenePos), lightColor[1]);
        }
        #endif
    }

    #ifdef fogVolumeEnabled
        mat2x3 fogData    = mat2x3(stex(colortex7).rgb, stex(colortex8).rgb);
        BilateralFog(fogData[0], fogData[1]);
        applyFogData(sceneColor.rgb, fogData);
    #endif

    if (isEyeInWater == 1) {
        sceneColor  = waterFog(sceneColor, length(scenePos), lightColor[1]);
    }

    bloomTiles    = vec4(0.0);

    //sceneColor.rgb  = texture(colortex12, 1-coord).rgb;
}