#version 430 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


/* RENDERTARGETS: 8 */
layout(location = 0) out vec4 screenLighting;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/colorspace.glsl"

in vec2 coord;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;

uniform sampler2D depthtex0;

uniform sampler2D noisetex;

uniform int frameCounter;

uniform float near, far;

uniform vec2 taaOffset;
uniform vec2 pixelSize, viewSize;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

/* ------ includes ------*/
#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

#define FUTIL_MAT16
#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"


vec4 packScreenLighting(vec3 directLight, vec3 emission) {
    vec4 lightRGBE  = encodeRGBE8(directLight);
    vec4 emissionRGBE = encodeRGBE8(emission);

    return vec4(pack2x8(lightRGBE.xy),
                pack2x8(lightRGBE.zw),
                pack2x8(emissionRGBE.xy),
                pack2x8(emissionRGBE.zw));
}

/* ------ BRDF ------ */

#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/hammon.glsl"
#include "/lib/brdf/labPBR.glsl"

#include "/lib/mat/emission.glsl"

void main() {
    float sceneDepth = stex(depthtex0).x;
    vec3 emission   = vec3(0.0);

    if (landMask(sceneDepth)) {
        ivec2 UV        = ivec2(coord * viewSize);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        materialLAB material = decodeSpecularTexture(Gbuffer1.MaterialData);

        vec3 albedo     = stex(colortex0).rgb;

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum));

        #if labEmissionMode != 0
        emission        = getEmissionScreenSpace(Gbuffer0.BlockID.y, albedo, material.emission);
        #else
        emission        = getEmissionScreenSpace(Gbuffer0.BlockID.y, albedo, 0.0);
        #endif
    }

    screenLighting      = packScreenLighting(vec3(0), emission);
}