#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


/* RENDERTARGETS: 0,3,5 */
layout(location = 0) out vec4 sceneColor;
layout(location = 1) out vec4 translucency;
layout(location = 2) out vec4 screenLighting;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat2x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex8;

uniform sampler2D depthtex0;
uniform sampler2D depthtex2;

uniform sampler2D noisetex;

uniform int frameCounter;

uniform float aspectRatio;
uniform float lightFlip;
uniform float sunAngle;
uniform float near, far;

uniform vec2 pixelSize, viewSize;

uniform mat4 gbufferProjection;

#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"

#include "/lib/light/sh.glsl"

/* ------ ATROUS ------ */

#include "/lib/offset/gauss.glsl"

#define atrousDepthExp 2.0

#define atrousNormalExponent 16

vec4 FetchUnpackedGbuffer(ivec2 UV) {
    vec4 Val  = texelFetch(colortex5, UV, 0);
    return vec4(Val.rgb * 2.0 - 1.0, sqr(Val.w));
}
SphericalHarmonics FetchSH(ivec2 uv) {
    return SphericalHarmonics(texelFetch(colortex6, uv, 0), texelFetch(colortex7, uv, 0).xy);
}

SphericalHarmonics spatialUpscaleSH(vec3 sceneNormal, float centerDepth) {
    ivec2 pos           = ivec2(coord * viewSize * indirectResScale);

    vec4 centerData     = vec4(sceneNormal, (depthLinear(centerDepth)));

    SphericalHarmonics total = InitSH();
    float totalWeight   = 0.0;

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos      = kernelO_5x5[i];

        ivec2 UV            = pos + deltaPos;

        bool valid          = all(greaterThanEqual(UV, ivec2(0))) && all(lessThan(UV, ivec2(viewSize)));

        if (!valid) continue;

        vec4 currentData    = FetchUnpackedGbuffer(UV);

        float depthDelta    = abs(currentData.w - centerData.w) * far * atrousDepthExp;

        float spatialWeight = pow(max(0.0, dot(currentData.xyz, centerData.xyz)), atrousNormalExponent);
            spatialWeight  *= exp(-depthDelta) * kernelW_5x5[i];

        //if (weight < 1e-20) continue;

        //accumulate stuff
        accumulateSH(total, FetchSH(UV), spatialWeight);

        totalWeight += spatialWeight;
    }

    //compensate for total sampling weight
    divideSH(total, max(totalWeight, 1e-16));

    return total;
}

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}
vec4 packDoubleRGBE8(vec3 colorA, vec3 colorB) {
    vec4 RGBE_A     = encodeRGBE8(colorA);
    vec4 RGBE_B     = encodeRGBE8(colorB);

    return vec4(pack2x8(RGBE_A.xy),
                pack2x8(RGBE_A.zw),
                pack2x8(RGBE_B.xy),
                pack2x8(RGBE_B.zw));
}

vec2 offsetDist(float x) {
	float n = fract(x * 8.0) * pi;
    return vec2(cos(n), sin(n)) * x;
}

float getSSAO(sampler2D depthtex, float depth, vec2 coord, float dither) {
    const uint steps = 3;
    const float baseRadius = sqrt2;

    bool hand       = depth < 0.56;
        depth       = depthLinear(depth);

    float currStep  = 0.2 * dither + 0.2;
	float fovScale  = gbufferProjection[1][1] / 1.37;
	float distScale = max((far - near) * depth + near, 5.0);
	vec2 scale      = baseRadius * vec2(1.0 / aspectRatio, 1.0) * fovScale / distScale;

    float ao = 0.0;

    const float maxOcclusionDist    = tau;
    const float anibleedExp         = 0.71;

    for (uint i = 0; i < steps; ++i) {
		vec2 offset = offsetDist(currStep) * scale;
		float mult  = (0.7 / baseRadius) * (far - near) * (hand ? 1024.0 : 1.0);

		float sampleDepth = depthLinear(texture(depthtex, coord + offset).r);
		float sample0 = (depth - sampleDepth) * mult;
        float antiBleed = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
            antiBleed   = saturate(antiBleed);
		float angle = mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
		float dist  = mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);

		sampleDepth = depthLinear(texture(depthtex, coord - offset).r);
		sample0     = (depth - sampleDepth) * mult;
        antiBleed   = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
           antiBleed   = saturate(antiBleed);
        angle      += mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
        dist       += mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);
		
		ao         += (clamp(angle + dist, 0.0, 1.0));
		currStep   += 0.2;
    }
	ao *= 1.0 / float(steps);

    if (saturate(ao) != ao) ao = 1.0;
	
	return sqrt(saturate(ao));
}
void main() {
    sceneColor          = stex(colortex0);
    float sceneDepth    = stex(depthtex0).x;
    vec3 albedo         = vec3(1.0);
    vec3 directLight    = vec3(1.0);
    //indirectLuma        = 0.0;
    //temporalAux         = stex(colortex14);

    if (landMask(sceneDepth)) {
        ivec2 UV        = ivec2(coord * viewSize);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        bool hand   = sceneDepth < stex(depthtex2).x;

            albedo          = sceneColor.rgb;
            
        mat2x3 screenLightingIn = unpackDoubleRGBE8(texelFetch(colortex8, UV, 0));

        vec3 emission       = screenLightingIn[1] / pi;
            if (hand) emission /= sqrPi;

        ivec2 IndirectUV    = ivec2(coord * viewSize * indirectResScale);

        #if indirectResReduction == 1
        SphericalHarmonics SHIn = FetchSH(UV);
        #else
        SphericalHarmonics SHIn = spatialUpscaleSH(Gbuffer0.SceneNormal, sceneDepth);
        #endif

        vec3 PTLight  = ProjectIrradiance(SHIn, Gbuffer0.SceneNormal);
            PTLight  += lightColor[0] / tau;
        #ifdef textureAOEnabled
            PTLight  *= Gbuffer1.OcclusionData.x;
        #endif

        #ifdef ssaoLayerEnabled
            PTLight  *= getSSAO(depthtex0, sceneDepth, coord, ditherBluenoise());
        #endif

        sceneColor.rgb     *= PTLight + emission;

        //sceneColor.rgb  = emission;

        screenLighting  = packDoubleRGBE8(clamp16F(screenLightingIn[0]), clamp16F(albedo));
    }

    //sceneColor.rgb  = albedo*0 + voxelDebug(sceneDepth) * 4;

    sceneColor          = drawbufferClamp(sceneColor);
    translucency        = vec4(0.0);
}