#version 450 compatibility

/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 0,3 */
layout(location = 0) out vec3 sceneColor;
layout(location = 1) out vec4 bloomTiles;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex8;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;

uniform int frameCounter;
uniform int isEyeInWater;
uniform int worldTime;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float lightFlip;
uniform float sunAngle;
uniform float wetness, rainStrength;

uniform vec2 skyCaptureResolution;
uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 lightDir, lightDirView;

uniform vec4 daytime;

uniform ivec2 eyeBrightness;
uniform ivec2 eyeBrightnessSmooth;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/atmos/waterConst.glsl"

#include "/lib/atmos/phase.glsl"
#include "/lib/atmos/air/const.glsl"
#include "/lib/atmos/project.glsl"

#include "/lib/atmos/fog.glsl"

void applyFogData(inout vec3 color, in mat2x3 data) {
    color = color * data[1] + data[0];
}

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

/* REFLECTIONS */
#include "/lib/util/colorspace.glsl"
#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/material.glsl"
#include "/lib/brdf/specular.glsl"
#include "/lib/frag/capture.glsl"

mat2x3 unpackReflectionAux(vec4 data){
    vec3 shadows    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 albedo     = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(shadows, albedo);
}

/*
    Based on the bilateral blur chocapic13 does.
    Actually seems illegally more effective than my checkerboard solution. Can't make this shit up...
*/

#define FOG_SCATTER_IMAGE colortex7
#define FOG_TRANSMITTANCE_IMAGE colortex8

ivec2 clampUV_Fog(ivec2 UV) {
    return clamp(UV, ivec2(0), ivec2(viewSize / 2 - 1));
}

void BilateralFog(out vec3 scattering, out vec3 transmittance) {
    scattering  = vec3(0);
    transmittance = vec3(0);

    vec3 zmult      = vec3(1.0/(far * near),far+near,far-near);

    #define OFFSET_SIZE 2

    ivec2 depthUV   = ivec2(coord * viewSize);
    ivec2 fogUV     = ivec2(coord * viewSize / FOG_LOWRES_FACTOR);
    ivec2 jitter    = (ivec2(gl_FragCoord.xy + frameCounter) % OFFSET_SIZE) * 2;

    float depthCenter = depthLinear(texelFetch(depthtex0, depthUV, 0).x);

    float sumWeight = 0.0;

    ivec2 depthTapUV = depthUV + ivec2(-OFFSET_SIZE) + jitter;
    float depthTap  = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
    float weight    = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
    ivec2 fogTapUV  = clampUV_Fog(fogUV + ivec2(-OFFSET_SIZE) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR);
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(-OFFSET_SIZE, 0) + jitter;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = clampUV_Fog(fogUV + ivec2(-OFFSET_SIZE, 0) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR);
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(0) + jitter;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = clampUV_Fog(fogUV + ivec2(0) + jitter / FOG_LOWRES_FACTOR);
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(0, -OFFSET_SIZE) + jitter;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = clampUV_Fog(fogUV + ivec2(0, -OFFSET_SIZE) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR);
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

    scattering     /= sumWeight;
    transmittance  /= sumWeight;

    return;
}

#define colorSampler colortex3
#define fresnelSampler colortex6

vec4 FetchGbuffer(ivec2 UV) {
        UV          = ivec2(vec2(UV) / reflectionResScale);
    vec3 normals    = decodeNormal(texelFetch(colortex1, UV, 0).xy);
    float depth     = depthLinear(texelFetch(depthtex0, UV, 0).x) * far;
    return vec4(normals, depth);
}

float sphericalGaussian(vec3 dir1, vec3 dir2, float lambda) {
    float w = exp(lambda * (saturate(dot(dir1, dir2)) - 1.0));
    return w > 1e-8 ? w : exp(-saturate(1-dot(dir1, dir2)) * 16.0);
}

float lobeSharpness(float roughness) {
    return 2.0 / sqr(max(roughness, 1e-5));
}

float specularLobeWeight(vec3 centerNorm, vec3 norm, float centerRoughness, float roughness, float beta) {
    float lambaCenter = lobeSharpness(centerRoughness);
    float lambaSample = lobeSharpness(roughness);
    return pow(2.0 * sqrt(lambaCenter*lambaSample) / max(lambaCenter + lambaSample, 1e-5), beta) 
        * sphericalGaussian(centerNorm, norm, beta * lambaCenter * lambaSample / max(lambaCenter + lambaSample, 1e-5));
}

const float lobeFactor = SPF_LOBE;
const float sharpenWeight = SPF_SHARPEN;

vec4 modifySpecularColor(vec4 color) {
    return vec4(pow(length(color.rgb), sharpenWeight) * normalizeSafe(color.rgb), color.a);
}
vec4 inverseModifySpecularColor(vec4 color) {
    return vec4(pow(length(color.rgb), 1.0 / sharpenWeight) * normalizeSafe(color.rgb), saturate(color.a));
}

uniform sampler3D depthtex2;
vec3 ditherBluenoiseF3() {
    ivec2 coord = ivec2(gl_FragCoord.xy);

    vec3 noise  = texelFetch(depthtex2, ivec3(coord & 255, 0), 0).xyz;
        noise   = fract(noise + torroidalShift.xyz * frameCounter);

    return noise;
}


void specularSpatialUpscale(out vec4 reflection, out vec3 fresnel, vec2 uv, vec4 centerData, float centerRoughness) {
    reflection  = vec4(0);
    fresnel     = vec3(0);

    ivec2 UV    = ivec2(uv * viewSize * reflectionResScale);

    //reflection = texelFetch(colorSampler, UV, 0);
    //fresnel = texelFetch(fresnelSampler, UV, 0).rgb;
    //return;

    float totalWeight = 0.0;

    const float normalExp = 256.0;

    ivec2 jitter        = ivec2((ditherBluenoiseF3().xy - 0.5));

    const int r = 2;
	for(int y = -r; y <= r; ++y) {
		for(int x = -r; x <= r; ++x) {
			ivec2 p = UV + ivec2(x, y) + jitter;

            bool valid          = all(greaterThanEqual(p, ivec2(0))) && all(lessThan(p, ivec2(viewSize * reflectionResScale)));

            if (!valid) continue;

            vec4 currentData    = FetchGbuffer(p);

            vec4 currentColor   = texelFetch(colorSampler, p, 0);
            vec4 currentFresnel = texelFetch(fresnelSampler, p, 0);

            float w         = float(valid);
                //w          *= specularKernel[abs(x)][abs(y)];

            float distDepth = abs(centerData.a - currentData.a) * 4.0;

                //w *= pow(max(0.0, dot(centerData.rgb, currentData.rgb)), normalExp);
                w *= exp(-distDepth);
                w *= max0(specularLobeWeight(centerData.xyz, currentData.xyz, centerRoughness, currentFresnel.a, lobeFactor));
            
            //if (w < 1e-16) continue;

            //accumulate stuff
            reflection  += modifySpecularColor(currentColor) * w;
            fresnel     += currentFresnel.rgb * w;

            totalWeight += w;
        }
    }
    totalWeight     = max(totalWeight, 1e-16);

    reflection  /= totalWeight;
    reflection   = inverseModifySpecularColor(reflection);
    fresnel.rgb /= totalWeight;
}

/* ------ rainbows ------ */

#ifdef rainbowsEnabled

vec3 hsv2rgb(vec3 c) {
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

vec3 generateRainbow(vec3 viewDir, float falloff) {
    float d     = dot(viewDir, -lightDirView);

    float h1    = clamp(linStep(d, 0.4, 0.5), 0.0, 0.5);
    float f1    = linStep(d, 0.395, 0.41) * sqr(1.0 - linStep(d, 0.43, 0.46));
    float b1    = linStep(d, 0.4, 0.42) * sqr(1.0 - linStep(d, 0.46, 0.75));

    vec3 c1     = hsv2rgb(vec3(h1, 1.0, 1.0)) * f1 + b1 * 0.3;
        c1     *= falloff;

    d  += 0.15;
    float h2    = clamp(0.5 - linStep(d, 0.35, 0.5), 0.0, 0.5);
    float f2    = linStep(d, 0.345, 0.38) * (1.0 - linStep(d, 0.42, 0.44));

    vec3 c2     = hsv2rgb(vec3(h2, 1.0, 1.0)) * f2 * 0.4 * sqr(falloff);

    vec3 lcol   = sunAngle > 0.5 ? lightColor[2] : lightColor[0];

    return (c1 + c2) * lcol * 0.014 * lightFlip;
}

#endif

void main() {
    sceneColor = stex(colortex0).rgb;

    vec2 sceneDepth = vec2(stex(depthtex0).x, 1.0);

    vec3 viewPos0   = screenToViewSpace(vec3(coord, sceneDepth.x));
    vec3 scenePos0  = viewToSceneSpace(viewPos0);

    vec3 viewDir    = normalize(viewPos0);
    vec3 worldDir   = normalize(scenePos0);

    float vDotL     = dot(normalize(viewPos0), lightDirView);

    float caveMult  = sqr(linStep(eyeBrightnessSmooth.y/240.0, 0.1, 0.9));

    float bluenoise     = ditherBluenoise();

    vec3 fogcol     = (sunAngle<0.5 ? lightColor[0] : lightColor[1]) * lightFlip * mieHG(vDotL, 0.68) * sqr(caveMult);
        fogcol     *= mix(0.33, 1.0, sqrt(abs(lightDir.y)));
        fogcol     += lightColor[2] * caveMult * rcp(pi);

    if (landMask(sceneDepth.x)) {
        ivec2 UV        = ivec2(coord * viewSize);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        vec3 viewNormal = mat3(gbufferModelView) * Gbuffer0.SceneNormal;

        bool water      = Gbuffer0.BlockID.x == 102;

        mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, ivec2(coord * viewSize), 0));

        vec3 viewDir    = normalize(viewPos0);

        mat2x3 reflectionAux = lightingData;
            reflectionAux[0] *= pi;

        vec3 directCol  = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * lightFlip;
            reflectionAux[0] = min(reflectionAux[0], directCol);

        materialProperties material = decodeLabBasic(Gbuffer1.MaterialData.xy);
        if (water) material = materialProperties(0.0001, 0.02, false, false, mat2x3(0.0));

        if (dot(viewDir, viewNormal) > 0.0) viewNormal = -viewNormal;

        vec3 reflectDir = reflect(viewDir, viewNormal);
        
        vec4 reflection = texture(colortex3, coord * reflectionResScale);
        vec3 fresnel    = texture(colortex6, coord * reflectionResScale).rgb;

        specularSpatialUpscale(reflection, fresnel, coord, vec4(Gbuffer0.SceneNormal, depthLinear(sceneDepth.x) * far), material.roughness);

        if (material.conductorComplex && material.conductor) fresnel *= sqrt(reflectionAux[1]) * pi;

        #ifdef resourcepackReflectionsEnabled
            //if (material.conductor) {
            //    float lumratio  = getLuma(reflection.rgb) / max(getLuma(sceneColor.rgb), 1e-8);
            //    reflection.rgb += vec3(getLuma(sceneColor.rgb)) * (1.0 - sstep(lumratio, 0.1, 0.3)) * rpi * (1.0 - sstep(getLuma(reflection.rgb), 0.01, 0.1));
           // }

            if (material.conductor) sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb * fresnel, reflection.a);
            else sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb, fresnel * reflection.a);            
        #else
            /* --- WATER REFLECTIONS --- */
            if (water) {
                sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb, fresnel * reflection.a);
            }
        #endif

        if (water) sceneColor.rgb += specularBeckmann(-viewDir, lightDirView, viewNormal, material) * reflectionAux[0];
        #ifdef specularHighlightsEnabled
        else sceneColor.rgb += specularTrowbridgeReitzGGX(-viewDir, lightDirView, viewNormal, material, reflectionAux[1]) * reflectionAux[0];
        #endif

        //sceneColor.rgb = reflection.rgb;

        //sceneColor.rgb = vec3(sqrt(material.roughness) > 0.25);
    }

    #ifndef waterVolumeEnabled
        if (isEyeInWater==1) sceneColor.rgb = waterFog(sceneColor.rgb, length(scenePos0), fogcol);
    #endif

    #ifdef fogVolumeEnabled
        mat2x3 fogData    = mat2x3(stex(colortex7).rgb, stex(colortex8).rgb);
        BilateralFog(fogData[0], fogData[1]);
        applyFogData(sceneColor.rgb, fogData);
    #endif

    #ifdef rainbowsEnabled
        float dist      = length(scenePos0);
        float falloff   = landMask(sceneDepth.x) ? sstep(dist, 128.0, 192.0) : 1.0;
        float fade      = saturate((wetness - rainStrength) * 1.33 - 0.02);

        if (falloff > 0.0 && fade > 0.0) {
            sceneColor.rgb  += generateRainbow(normalize(viewPos0), falloff) * fade;
        }
    #endif

    sceneColor          = drawbufferClamp(sceneColor);
    bloomTiles          = vec4(0);
}