#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


/* RENDERTARGETS: 0,3,5 */
layout(location = 0) out vec4 sceneColor;
layout(location = 1) out vec4 translucency;
layout(location = 2) out vec4 screenLighting;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex8;

uniform sampler2D depthtex0;
uniform sampler2D depthtex2;

uniform sampler2D noisetex;

uniform int frameCounter;

uniform float aspectRatio;
uniform float lightFlip;
uniform float sunAngle;
uniform float near, far;
uniform float eyeAltitude;

uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;

uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferModelView, gbufferModelViewInverse;

#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/frag/bluenoise.glsl"

#include "/lib/light/sh.glsl"

/* ------ ATROUS ------ */

#include "/lib/offset/gauss.glsl"

#define atrousDepthExp 2.0

#define atrousNormalExponent 16

vec4 FetchUnpackedGbuffer(ivec2 UV) {
    vec4 Val  = texelFetch(colortex5, UV, 0);
    return vec4(Val.rgb * 2.0 - 1.0, sqr(Val.w));
}
SphericalHarmonics FetchSH(ivec2 uv) {
    return SphericalHarmonics(texelFetch(colortex6, uv, 0), texelFetch(colortex7, uv, 0).xy);
}

SphericalHarmonics spatialUpscaleSH(vec3 sceneNormal, float centerDepth) {
    ivec2 pos           = ivec2(coord * viewSize * indirectResScale);

    vec4 centerData     = vec4(sceneNormal, (depthLinear(centerDepth)));

    SphericalHarmonics total = InitSH();
    float totalWeight   = 0.0;

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos      = kernelO_5x5[i];

        ivec2 UV            = pos + deltaPos;

        bool valid          = all(greaterThanEqual(UV, ivec2(0))) && all(lessThan(UV, ivec2(viewSize)));

        if (!valid) continue;

        vec4 currentData    = FetchUnpackedGbuffer(UV);

        float depthDelta    = abs(currentData.w - centerData.w) * far * atrousDepthExp;

        float spatialWeight = pow(max(0.0, dot(currentData.xyz, centerData.xyz)), atrousNormalExponent);
            spatialWeight  *= exp(-depthDelta) * kernelW_5x5[i];

        //if (weight < 1e-20) continue;

        //accumulate stuff
        accumulateSH(total, FetchSH(UV), spatialWeight);

        totalWeight += spatialWeight;
    }

    //compensate for total sampling weight
    divideSH(total, max(totalWeight, 1e-16));

    return total;
}

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}
vec4 packDoubleRGBE8(vec3 colorA, vec3 colorB) {
    vec4 RGBE_A     = encodeRGBE8(colorA);
    vec4 RGBE_B     = encodeRGBE8(colorB);

    return vec4(pack2x8(RGBE_A.xy),
                pack2x8(RGBE_A.zw),
                pack2x8(RGBE_B.xy),
                pack2x8(RGBE_B.zw));
}

vec2 offsetDist(float x) {
	float n = fract(x * 8.0) * pi;
    return vec2(cos(n), sin(n)) * x;
}

float getSSAO(sampler2D depthtex, float depth, vec2 coord, float dither) {
    const uint steps = 3;
    const float baseRadius = sqrt2;

    bool hand       = depth < 0.56;
        depth       = depthLinear(depth);

    float currStep  = 0.2 * dither + 0.2;
	float fovScale  = gbufferProjection[1][1] / 1.37;
	float distScale = max((far - near) * depth + near, 5.0);
	vec2 scale      = baseRadius * vec2(1.0 / aspectRatio, 1.0) * fovScale / distScale;

    float ao = 0.0;

    const float maxOcclusionDist    = pi;
    const float anibleedExp         = 0.71;

    for (uint i = 0; i < steps; ++i) {
		vec2 offset = offsetDist(currStep) * scale;
		float mult  = (0.7 / baseRadius) * (far - near) * (hand ? 1024.0 : 1.0);

		float sampleDepth = depthLinear(texture(depthtex, coord + offset).r);
		float sample0 = (depth - sampleDepth) * mult;
        float antiBleed = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
            antiBleed   = saturate(antiBleed);
		float angle = mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
		float dist  = mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);

		sampleDepth = depthLinear(texture(depthtex, coord - offset).r);
		sample0     = (depth - sampleDepth) * mult;
        antiBleed   = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
           antiBleed   = saturate(antiBleed);
        angle      += mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
        dist       += mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);
		
		ao         += (clamp(angle + dist, 0.0, 1.0));
		currStep   += 0.2;
    }
	ao *= 1.0 / float(steps);

    if (saturate(ao) != ao) ao = 1.0;
	
	return sqrt(saturate(ao)) * 0.9 + 0.1;
}

/* --- RTAO --- */

#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/hammon.glsl"

vec3 screenspaceRT(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 10;
    const float stepSize    = halfPi / float(maxSteps);

    vec3 stepVector         = direction * stepSize;

    vec3 endPosition        = position + stepVector * maxSteps;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec2 maxPosXY           = max(abs(endScreenPosition.xy * 2.0 - 1.0), vec2(1.0));
    float stepMult          = minOf(vec2(1.0) / maxPosXY);
        stepVector         *= stepMult;

    // closest texel iteration
    vec3 samplePos          = position;
    
        samplePos          += stepVector / 6.0;
    vec3 screenPos          = viewToScreenSpace(samplePos);

    const float thresh      = 0.25;
/*
    if (saturate(screenPos.xy) == screenPos.xy) {
        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }*/
    
        samplePos          += stepVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        vec3 screenPos      = viewToScreenSpace(samplePos);
            samplePos      += stepVector;
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }

    return vec3(1.1);
}

vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}
vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}

#define HASHSCALE1 .1031
float hash12(vec2 x) {
    vec3 x3  = fract(vec3(x.xyx) * HASHSCALE1);
        x3  += dot(x3, x3.yzx + 19.19);
    return fract((x3.x + x3.y) * x3.z);
}

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}

float ditherHashNoise() {
    float noise     = hash12(gl_FragCoord.xy);
        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)) *0);

    return noise;
}
float ditherBluenoise2() {
    ivec2 coord = ivec2(fract(gl_FragCoord.xy/256.0)*256.0);
    float noise = texelFetch(noisetex, coord, 0).a;

        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)));

    return noise;
}


#define RTAO_SSP 3

float getSSRTAO(vec3 viewPos, vec3 sceneNormal, vec2 dither, bool hand) {
    vec3 viewNormal     = mat3(gbufferModelView) * sceneNormal;

    float occlusion     = 0.0;

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    if (hand) return 1.0;

    /*
        Single Bounce 
    */

    vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2) * frameCounter + 0.5);

    vec2 noiseCurr      = hash22(gl_FragCoord.xy + frameCounter);

    for (uint i = 0; i < RTAO_SSP; ++i) {
        ++quasirandomCurr;
        noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));

        vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

        vec3 rayDirection   = GenerateCosineVectorSafe(sceneNormal, vectorXY);
            rayDirection    = normalize(mat3(gbufferModelView) * rayDirection);

        if (dot(viewNormal, rayDirection) < 0.0) rayDirection = -rayDirection;

        vec3 hitPosition    = screenspaceRT(viewPos, rayDirection, dither.y);
        //float brdf          = saturate(dot(rayDirection, viewNormal)) * sqrt2;

        if (!(hitPosition.z < 1.0)) {
            occlusion  += 1.0;
        }
    }

    occlusion      /= float(RTAO_SSP);

    return saturate(occlusion);
}

#if DEBUG_VIEW == 4

uniform sampler2D colortex15;

#include "/lib/light/cacheLib.glsl"

vec3 cacheDebug(vec3 position, vec3 sceneNormal) {

    position    = screenToViewSpace(position);
    position    = viewToSceneSpace(position);

    int face = getFace(sceneNormal);
    return cacheRead(getCachePixel(position - sceneNormal * 0.01, face)).rgb;
    //return sampleCacheHQ(position, sceneNormal, face);
}
#endif


void main() {
    sceneColor          = stex(colortex0);
    float sceneDepth    = stex(depthtex0).x;
    vec3 albedo         = vec3(1.0);
    vec3 directLight    = vec3(1.0);
    //indirectLuma        = 0.0;
    //temporalAux         = stex(colortex14);

    if (landMask(sceneDepth)) {
        ivec2 UV        = ivec2(coord * viewSize);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        bool hand   = sceneDepth < stex(depthtex2).x;

            albedo          = sceneColor.rgb;

        #if DEBUG_VIEW == 1
            sceneColor.rgb  = vec3(1.0);
        #endif

        mat2x3 screenLightingIn = unpackDoubleRGBE8(texelFetch(colortex8, UV, 0));

        vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * lightFlip;
            screenLightingIn[0] *= directCol;

        vec3 emission       = screenLightingIn[1] / pi;
            if (hand) emission /= sqrPi;

        #ifndef boostEmitters
            emission       *= rpi*0;
        #endif

        ivec2 IndirectUV    = ivec2(coord * viewSize * indirectResScale);

        #if indirectResReduction == 1
        SphericalHarmonics SHIn = FetchSH(UV);
        #else
        SphericalHarmonics SHIn = spatialUpscaleSH(Gbuffer0.SceneNormal, sceneDepth);
        #endif

        vec3 PTLight  = ProjectIrradiance(SHIn, Gbuffer0.SceneNormal);
            PTLight  += minimumAmbientColor * minimumAmbientMult * minimumAmbientIllum;
        #ifdef textureAOEnabled
            PTLight  *= Gbuffer1.OcclusionData.x;
        #endif

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth));
            
        #ifdef ssaoLayerEnabled
            //float ssao  = getSSRTAO(viewPos, Gbuffer0.SceneNormal, vec2(ditherBluenoise2(), ditherHashNoise()), hand);
            float ssao  = getSSAO(depthtex0, sceneDepth, coord, ditherBluenoise());
            PTLight  *= ssao;
        #endif

        sceneColor.rgb     *= screenLightingIn[0] + PTLight + emission;

        #if DEBUG_VIEW == 2
            sceneColor.rgb      = PTLight;
        #endif

        #if DEBUG_VIEW == 3
            sceneColor.rgb      = albedo;
        #endif

        //sceneColor.rgb  = vec3(ssao);

        #if DEBUG_VIEW == 4
            sceneColor.rgb  = cacheDebug(vec3(coord, sceneDepth), Gbuffer0.SceneNormal);
        #endif

        screenLighting  = packDoubleRGBE8(clamp16F(screenLightingIn[0]), clamp16F(albedo));
    }

    //sceneColor.rgb  = albedo*0 + voxelDebug(sceneDepth) * 4;

    sceneColor          = drawbufferClamp(sceneColor);
    translucency        = vec4(0.0);
}