#version 450 compatibility

/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 0,6,7 */
layout(location = 0) out vec3 sceneColor;
layout(location = 1) out vec3 fogScattering;
layout(location = 2) out vec3 fogTransmittance;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;

uniform int frameCounter;
uniform int isEyeInWater;
uniform int worldTime;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float lightFlip;
uniform float sunAngle;
uniform float wetness;
uniform float worldAnimTime;

uniform vec2 skyCaptureResolution;
uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 lightDir, lightDirView;

uniform vec4 daytime;

uniform ivec2 eyeBrightness;
uniform ivec2 eyeBrightnessSmooth;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

#define GBufferREAD0
#include "/lib/gbuffer.glsl"

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/atmos/waterConst.glsl"

#include "/lib/light/cacheLib.glsl"

vec3 blendTranslucencies(vec3 sceneColor, vec4 translucents, vec3 albedo) {
    vec3 color  = sceneColor;

    #ifdef vibrantAbsorptionEnabled
        color  *= albedo;
    #else
        color  *= mix(vec3(1.0), albedo, sqrt(translucents.a));
    #endif
    
        color   = mix(color, translucents.rgb, translucents.a);

    return color;
}

#include "/lib/light/warp.glsl"

vec3 getShadowmapPos(vec3 scenePos) {  //shadow 2d
    vec3 pos    = scenePos + vec3(0.08)*lightDir;
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;

        pos.xy  = shadowmapWarp(pos.xy);

        pos.x   = pos.x * 0.5 + 0.5;
        pos.xy  = pos.yx;

    return pos*0.5+0.5;
}

float readCloudShadowmap(sampler2D shadowmap, vec3 position) {
    position    = mat3(shadowModelView) * position;
    position   /= cloudShadowmapRenderDistance;
    position.xy = position.xy * 0.5 + 0.5;

    position.xy /= vec2(1.0, 1.0 + (1.0 / 3.0));

    return texture(shadowmap, position.xy).a;
}

vec3 shadowColorSample(sampler2D tex, vec2 position) {
    vec4 colorSample = texture(shadowcolor0, position);
    return mix(vec3(1.0), colorSample.rgb * 4.0, colorSample.a);
}

#include "/lib/atmos/phase.glsl"
#include "/lib/atmos/air/const.glsl"
#include "/lib/frag/noise.glsl"

uniform vec3 fogDensityCoeff;

#define fogMistAltitude 100.0   //[30.0 40.0 50.0 60.0 70.0 80.0 90.0 100.0 110.0 120.0 130.0 140.0 150.0]
#define fogSeaLevel 64.0    //[8.0 16.0 24.0 32.0 40.0 48.0 56.0 64.0 72.0 80.0 88.0 96.0]
#define fogMistFalloff 1.0  //[0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 1.1 1.2 1.3 1.4 1.5 1.6 1.7 1.8 1.9 2.0]
#define airDensity 1.0
#define fogMinSteps 10
#define fogAdaptiveSteps 12     //[4 6 8 10 12 14 16 18]
#define fogClipDist 512.0       //[256.0 384.0 448.0 512.0 768.0 1024.0]
#define fogMistLightingSteps 2  //[2 4 6 8 10 12]

#ifdef freezeAtmosAnim
    const float fogTime   = float(atmosAnimOffset) * 0.006;
#else
    #ifdef volumeWorldTimeAnim
        float fogTime     = worldAnimTime * 3.6;
    #else
        float fogTime     = frameTimeCounter * 0.006;
    #endif
#endif

vec3 fogAirDensity(vec3 rPos, float altitude) {
    float maxFade = sqr(1.0 - linStep(altitude, 144.0, 256.0));

    vec2 rm     = expf(-max0((altitude - fogSeaLevel) * fogFalloffScale));
        rm     *= fogAirScale * fogDensityCoeff.xy * maxFade * fogDensityMult;

    float mist  = 0.0;

    if (fogDensityCoeff.z > 1e-8) {
        vec3 wind   = vec3(fogTime, 0.0, fogTime * 0.5);
        float noise = value3D(rPos * 0.03 + wind) + value3D(rPos * 0.06 + wind * pi) * 0.5;

            mist  = expf(-max0((altitude - fogSeaLevel) * 0.01 * fogMistFalloff));
            mist    = sstep(noise * rcp(1.5) - (1.0 - mist) * 0.5, 0.4, 0.7);
            mist   *= sqr(1.0 - linStep(altitude, fogMistAltitude - 15.0, fogMistAltitude));

        mist *= fogDensityCoeff.z;
    }

    return vec3(rm, mist);
}

vec2 airPhaseFunction(float cosTheta) {
    return vec2(rayleighPhase(cosTheta), mieCS(cosTheta, airMieG));
}
float airMieBackscatter(float cosTheta) {
    return mieHG(cosTheta, -airMieG * rcp(pi));
}
float airMieBackscatter(float cosTheta, float g) {
    return mieHG(cosTheta, -g * rcp(pi));
}
float airPhaseMist(float cosTheta, float g) {
    g  *= mistMieAnisotropy;
    return mix(mieCS(cosTheta, g), mieHG(cosTheta, -g * rpi), mistMieAnisotropy / 4.0);
}
float fogMistPhase(float cosTheta, float density) {
    return mix(mieCS(cosTheta, pow(airMieG, 1.0 + density)), airMieBackscatter(cosTheta, airMieG), 0.19);
}

vec3 readVolumeLight(ivec3 index) {
    if (outsideCacheVolume(index)) return vec3(0);
    return clamp16F(texelFetch(colortex15, getCachePixelCacheSpace(index, 6), 0).rgb);
}
vec3 readVolumeLight(ivec3 index, vec3 skylight) {
    if (outsideCacheVolume(index)) return skylight;
    return clamp16F(texelFetch(colortex15, getCachePixelCacheSpace(index, 6), 0).rgb);
}

vec3 readVolumeLightSmooth(vec3 scenePos, vec3 skylight) {
    ivec3 voxelIndex = sceneToCacheSpace(scenePos);

    #ifdef radianceVolumeSmoothRead
        vec3 alpha  = fract(scenePos + cameraPosition - 0.5);
            alpha   = linStep(alpha, 0.001, 0.999);

        vec3 direction = fract(scenePos + cameraPosition) - 0.5;
        bvec3 isPositive = greaterThan(direction, vec3(0));


        /*
            |    | Y+ |    |

            | X- | M  | X+ |

            |    | Y- |    |
        */

        ivec3 mid   = voxelIndex;
        ivec3 edge  = voxelIndex + ivec3(isPositive.x ? 1 : -1, isPositive.y ? 1 : -1, isPositive.z ? 1 : -1);

        alpha.x         = isPositive.x ? alpha.x : 1.0 - alpha.x;
        alpha.y         = isPositive.y ? alpha.y : 1.0 - alpha.y;
        alpha.z         = isPositive.z ? alpha.z : 1.0 - alpha.z;

        vec3 tap_LY_LZ  = mix(readVolumeLight(ivec3(mid.x, mid.y, mid.z), skylight), 
                            readVolumeLight(ivec3(edge.x, mid.y, mid.z), skylight), alpha.x);
        vec3 tap_LY_HZ  = mix(readVolumeLight(ivec3(mid.x, mid.y, edge.z), skylight), 
                            readVolumeLight(ivec3(edge.x, mid.y, edge.z), skylight), alpha.x);

        vec3 tap_HY_LZ  = mix(readVolumeLight(ivec3(mid.x, edge.y, mid.z), skylight), 
                            readVolumeLight(ivec3(edge.x, edge.y, mid.z), skylight), alpha.x);
        vec3 tap_HY_HZ  = mix(readVolumeLight(ivec3(mid.x, edge.y, edge.z), skylight), 
                            readVolumeLight(ivec3(edge.x, edge.y, edge.z), skylight), alpha.x);

        vec3 tap_LZ     = mix(tap_LY_LZ, tap_HY_LZ, alpha.y);
        vec3 tap_HZ     = mix(tap_LY_HZ, tap_HY_HZ, alpha.y);

        return mix(tap_LZ, tap_HZ, alpha.z);
    #else
        return readVolumeLight(voxelIndex, skylight);
    #endif
}

mat2x3 volumetricFog(vec3 scenePos, vec3 sceneDir, bool isSky, float vDotL, float dither, float cave) {
    float topDist    = length(sceneDir * ((256.0 - eyeAltitude) * rcp(sceneDir.y)));
    float bottomDist = length(sceneDir * ((-32.0 - eyeAltitude) * rcp(sceneDir.y)));

    float volumeHull = sceneDir.y > 0.0 ? topDist : bottomDist;

    float endDist   = isSky ? min(volumeHull, fogClipDist) : length(scenePos);
    float startDist = eyeAltitude > 256.0 ? topDist : 1.0;

    vec3 startPos   = eyeAltitude > 256.0 ? sceneDir * startDist : vec3(0.0);
        startPos   += gbufferModelViewInverse[3].xyz;
    vec3 endPos     = isSky ? sceneDir * endDist : scenePos;

    float baseStep  = length(endPos - startPos);
    float stepCoeff = saturate(baseStep * rcp(clamp(far, 256.0, 512.0)));

    uint steps      = fogMinSteps + uint(stepCoeff * fogAdaptiveSteps);

    vec3 rStep      = (endPos - startPos) / float(steps);
    vec3 rPos       = startPos + rStep * dither;
    float rLength   = length(rStep);

    vec3 shadowStartPos = transMAD(shadowModelView, (startPos));
        shadowStartPos  = projMAD(shadowProjection, shadowStartPos);
        shadowStartPos.z *= 0.2;
    vec3 shadowEndPos   = transMAD(shadowModelView, (endPos));
        shadowEndPos    = projMAD(shadowProjection, shadowEndPos);
        shadowEndPos.z *= 0.2;

    vec3 shadowStep = (shadowEndPos - shadowStartPos) / float(steps);
    vec3 shadowPos  = shadowStartPos + shadowStep * dither;

    mat2x3 scattering = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    vec3 phase      = vec3(airPhaseFunction(vDotL), airPhaseMist(vDotL, airMieG));
    float phaseIso  = 0.25;

    vec3 sunlight       = (worldTime>23000 || worldTime<12900) ? lightColor[0] : lightColor[2];
        sunlight       *= lightFlip;
    vec3 skylight       = lightColor[1] * cave;

    for (uint i = 0; i < steps; ++i, rPos += rStep, shadowPos += shadowStep) {
        if (maxOf(transmittance) < 0.01) break;

        float altitude  = rPos.y + eyeAltitude;

        if (altitude > 256.0) continue;

        vec3 density    = fogAirDensity(rPos + cameraPosition, altitude);

        vec3 stepRho    = density * rLength;
        vec3 od         = fogExtinctMat * stepRho;

        vec3 stepT      = expf(-od);
        vec3 scatterInt = saturate((stepT - 1.0) * rcp(-max(od, 1e-16)));
        vec3 visScatter = transmittance * scatterInt;

        vec3 sunScatter = vec3(0.0);
        //vec3 skyScatter = vec3(0.0);

        vec3 currPhase  = phase;

        for (uint j = 0; j < 4; ++j) {
            float n     = float(j);
            float m     = pow(0.5, n);

            phase.y     = mieCS(vDotL, airMieG * m);
            if (density.z > 0.0) phase.z = airPhaseMist(vDotL, airMieG * m);

            sunScatter += fogScatterMat * (stepRho * phase) * visScatter * m;
        }

        //skyScatter     += fogScatterMat * (stepRho * phaseIso) * visScatter;

        vec3 shadowCoord = vec3(shadowmapWarp(shadowPos.xy), shadowPos.z);
            shadowCoord.x   = shadowCoord.x * 0.5 + 0.5;
            shadowCoord.xy  = shadowCoord.yx;
            shadowCoord     = shadowCoord * 0.5 + 0.5;

        float shadow0   = texture(shadowtex0, shadowCoord);

        float shadow    = 1.0;
        vec3 shadowCol  = vec3(1.0);
        
        if (shadow0 < 1.0) {
            shadow      = texture(shadowtex1, shadowCoord);

            if (abs(shadow - shadow0) > 0.1) {
                shadowCol   = shadowColorSample(shadowcolor0, shadowCoord.xy);
            }
        }
        #ifdef cloudShadowsEnabled
        shadow         *= readCloudShadowmap(colortex4, rPos);
        #endif

        //vec3 lightScatter   = cacheReadFog(getCachePixel(rPos, 6), skylight * phaseIso);
        vec3 lightScatter   = readVolumeLightSmooth(rPos, skylight * phaseIso);
            lightScatter    = lightScatter * (fogScatterMat * (stepRho) * visScatter);

        scattering[0]  += (sunScatter * shadowCol * transmittance) * shadow;
        scattering[1]  += lightScatter * transmittance;

        transmittance  *= stepT;
    }

    vec3 color  = scattering[0] * sunlight + scattering[1];

    if (color != color) {   //because NaNs on nVidia don't need a logic cause to happen
        color = vec3(0.0);
        transmittance = vec3(1.0);
    }

    return mat2x3(color, saturate(transmittance));
}

/* ------ WATER VOLUME ------ */
#define waterVolMinSteps 3
#define waterVolAdaptiveSteps 9     //[2 4 6 8 10 12]
#define waterVolClipDist 48.0       //[16.0 24.0 32.0 40.0 48.0 56.0 64.0]

mat2x3 volumetricWater(vec3 scenePos, vec3 startPos0, vec3 worldDir, bool isSky, float vDotL, float dither, float cave) {
    float endDist   = isSky ? waterVolClipDist : length(scenePos);

    vec3 startPos   = startPos0;
        //startPos   += gbufferModelViewInverse[3].xyz;
    vec3 endPos     = isSky ? worldDir * waterVolClipDist : scenePos;
        if (distance(startPos, endPos) > waterVolClipDist) endPos = waterVolClipDist * worldDir;

    float baseStep  = length(endPos - startPos);
    float stepCoeff = saturate(baseStep * rcp(waterVolClipDist));
        //if (isEyeInWater == 1) stepCoeff = 1.0;

    uint steps      = waterVolMinSteps + uint(stepCoeff * waterVolAdaptiveSteps);

    vec3 rStep      = (endPos - startPos) / float(steps);
    vec3 rPos       = startPos + rStep * dither;
    float rLength   = length(rStep);

    vec3 shadowStartPos = transMAD(shadowModelView, (startPos));
        shadowStartPos  = projMAD(shadowProjection, shadowStartPos);
        shadowStartPos.z *= 0.2;
    vec3 shadowEndPos   = transMAD(shadowModelView, (endPos));
        shadowEndPos    = projMAD(shadowProjection, shadowEndPos);
        shadowEndPos.z *= 0.2;

    vec3 shadowStep = (shadowEndPos - shadowStartPos) / float(steps);
    vec3 shadowPos  = shadowStartPos + shadowStep * dither;

    mat2x3 scattering = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    float phase     = mieHG(vDotL, 0.7);
    float phaseIso  = 0.25 * rcp(pi);
        phase       = mix(phase, phaseIso, 0.3);

    vec3 sunlight       = (worldTime>23000 || worldTime<12900) ? lightColor[0] : lightColor[2];
        sunlight       *= lightFlip;
    vec3 skylight       = lightColor[1] * cave * rpi;

    for (uint i = 0; i < steps; ++i, rPos += rStep, shadowPos += shadowStep) {
        if (distance(startPos, rPos) > waterVolClipDist) continue;
        if (maxOf(transmittance) < 0.01) break;

        float altitude  = rPos.y + eyeAltitude;

        if (altitude > 256.0) continue;

        float stepRho   = rLength * waterDensity;

        vec3 shadowCoord = vec3(shadowmapWarp(shadowPos.xy), shadowPos.z);
            shadowCoord.x   = shadowCoord.x * 0.5 + 0.5;
            shadowCoord.xy  = shadowCoord.yx;
            shadowCoord     = shadowCoord * 0.5 + 0.5;

        float shadow0   = texture(shadowtex0, shadowCoord);

        float shadow    = 1.0;
        vec3 shadowCol  = vec3(1.0);
        
        if (shadow0 < 1.0) {
            shadow      = texture(shadowtex1, shadowCoord);

            if (abs(shadow - shadow0) > 0.1) {
                shadowCol   = shadowColorSample(shadowcolor0, shadowCoord.xy);
            }
        }
        #ifdef cloudShadowsEnabled
        shadow         *= readCloudShadowmap(colortex4, rPos);
        #endif

        //vec3 lightScatter   = cacheReadFog(getCachePixel(rPos, 6), skylight);
        vec3 lightScatter   = readVolumeLightSmooth(rPos, skylight * phaseIso);
            lightScatter    = lightScatter;

        scattering[0]  += (shadowCol * transmittance) * shadow;
        scattering[1]  += lightScatter * transmittance;

        transmittance  *= expf(-waterAttenCoeff * stepRho);

    }

    vec3 attenMul   = (1.0 - expf(-waterAttenCoeff * rLength)) * rcp(waterAttenCoeff);

    scattering[0]  *= waterScatterCoeff * attenMul * phase;
    scattering[1]  *= waterScatterCoeff * attenMul;

    vec3 color  = scattering[0] * sunlight + scattering[1];

    if (color != color) {   //because NaNs on nVidia don't need a logic cause to happen
        color = vec3(0.0);
        transmittance = vec3(1.0);
    }

    return mat2x3(color, saturate(transmittance));
}

void applyFogData(inout vec3 color, in mat2x3 data) {
    color = color * data[1] + data[0];
}

#include "/lib/atmos/fog.glsl"

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

/* ------ REFRACTION ------ */
vec3 refract2(vec3 I, vec3 N, vec3 NF, float eta) {     //from spectrum by zombye
    float NoI = dot(N, I);
    float k = 1.0 - eta * eta * (1.0 - NoI * NoI);
    if (k < 0.0) {
        return vec3(0.0); // Total Internal Reflection
    } else {
        float sqrtk = sqrt(k);
        vec3 R = (eta * dot(NF, I) + sqrtk) * NF - (eta * NoI + sqrtk) * N;
        return normalize(R * sqrt(abs(NoI)) + eta * I);
    }
}


void main() {
    sceneColor = stex(colortex0).rgb;

    ivec2 texelUV   = ivec2(coord * viewSize);

    GData00 Gbuffer0 = ReadGbuffer0(texelUV);

    vec2 sceneDepth = vec2(stex(depthtex0).x, stex(depthtex1).x);

    vec3 viewPos0   = screenToViewSpace(vec3(coord, sceneDepth.x));
    vec3 scenePos0  = viewToSceneSpace(viewPos0);

    vec3 viewPos1   = screenToViewSpace(vec3(coord, sceneDepth.y));
    vec3 scenePos1  = viewToSceneSpace(viewPos1);

    vec3 viewDir    = normalize(viewPos0);
    vec3 worldDir   = normalize(scenePos0);

    float vDotL     = dot(viewDir, lightDirView);

    bool translucent    = sceneDepth.x < sceneDepth.y;

    float caveMult  = sqr(saturate(float(eyeBrightnessSmooth.y) / 240.0));

    #ifdef refractionEnabled
    if (translucent && Gbuffer0.BlockID.x != 100){
        vec3 viewNormal     = mat3(gbufferModelView) * Gbuffer0.SceneNormal;
        //vec3 flatNormal     = normalize(cross(dFdx(scenePos0), dFdy(scenePos0)));
        vec3 flatNormal     = decodeNormal(unpack2x8(texelFetch(colortex2, texelUV, 0).z));

        if (flatNormal != flatNormal) flatNormal = Gbuffer0.SceneNormal;
            flatNormal = clampDIR(flatNormal);

        vec3 flatViewNormal = mat3(gbufferModelView) * flatNormal;

        vec3 normalCorrected = dot(viewNormal, viewDir) > 0.0 ? -viewNormal : viewNormal;

        vec3 refractedDir   = refract2(normalize(viewPos1), normalCorrected, flatViewNormal, rcp(1.33));
        //vec3 refractedDir   = refract(normalize(viewPos1), normalCorrected, rcp(1.33));

        float refractedDist = distance(viewPos0, viewPos1);

        vec3 refractedPos   = viewPos1 + refractedDir * refractedDist;

        vec3 screenPos      = viewToScreenSpace(refractedPos);

        float distToEdge    = maxOf(abs(screenPos.xy * 2.0 - 1.0));
            distToEdge      = sqr(sstep(distToEdge, 0.7, 1.0));

            screenPos.xy    = mix(screenPos.xy, coord, distToEdge);

        //vec2 refractionDelta = coord - screenPos.xy;

        float sceneDepthNew = texture(depthtex1, screenPos.xy).x;

        if (sceneDepthNew > sceneDepth.x) {
            sceneDepth.y    = sceneDepthNew;
            viewPos1        = screenToViewSpace(vec3(screenPos.xy, sceneDepth.y));
            scenePos1       = viewToSceneSpace(viewPos1);

            sceneColor.rgb  = texture(colortex0, screenPos.xy).rgb;
        }
    }
    #endif

    float bluenoise     = ditherBluenoise();

    bool water      = Gbuffer0.BlockID.x == 102;

    mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, ivec2(coord * viewSize), 0));

    float dist0     = distance(scenePos0, gbufferModelViewInverse[3].xyz);
    float dist1     = distance(scenePos1, gbufferModelViewInverse[3].xyz);

    vec3 fogcol     = (sunAngle<0.5 ? lightColor[0] : lightColor[1]) * lightFlip * mieHG(vDotL, 0.68) * sqr(caveMult);
        fogcol     *= mix(0.33, 1.0, sqrt(abs(lightDir.y)));
        fogcol     += lightColor[2] * caveMult * rcp(pi);

    if (translucent){

        if (isEyeInWater==0) {
            if (water) {
                #ifdef waterVolumeEnabled

                mat2x3 waterVol     = volumetricWater(scenePos1, scenePos0, worldDir, false, vDotL, bluenoise, caveMult);
                applyFogData(sceneColor.rgb, waterVol);

                #else

                sceneColor.rgb = waterFog(sceneColor.rgb, dist1-dist0, lightColor[1]);

                #endif
            }
            else {
                sceneColor.rgb = simpleFog(sceneColor.rgb, dist1-dist0, fogcol);
            }
        }

        sceneColor.rgb = blendTranslucencies(sceneColor.rgb, stex(colortex3), lightingData[1]);
    }

    fogScattering = vec3(0);
    fogTransmittance = vec3(1);

    #ifdef fogVolumeEnabled
    const float fogLOD      = 2.0;

    vec2 fogCoord           = coord * FOG_LOWRES_FACTOR;

    if (clamp(fogCoord, -pixelSize * FOG_LOWRES_FACTOR, 1.0 + pixelSize * FOG_LOWRES_FACTOR) == fogCoord) {
        fogCoord         = saturate(fogCoord);

        float sceneDepth = texelFetch(depthtex0, ivec2(fogCoord * viewSize), 0).x;

        vec3 viewPos0   = screenToViewSpace(vec3(fogCoord, sceneDepth));
        vec3 scenePos0  = viewToSceneSpace(viewPos0);

        vec3 viewDir    = normalize(viewPos0);
        vec3 worldDir   = normalize(scenePos0);

        float vDotL     = dot(viewDir, lightDirView);
        bool isSky      = !landMask(sceneDepth);

        mat2x3 fogData  = mat2x3(vec3(0.0), vec3(1.0));

        if (isEyeInWater == 1) {
            #ifdef waterVolumeEnabled
            fogData = volumetricWater(scenePos0, gbufferModelViewInverse[3].xyz, worldDir, isSky, vDotL, bluenoise, caveMult);
            #endif
        } else {
            #ifdef fogVolumeEnabled
            fogData    = volumetricFog(scenePos0, worldDir, isSky, vDotL, bluenoise, caveMult);
            #endif
        }

        fogScattering = clamp16F(fogData[0]);
        fogTransmittance = saturate(fogData[1]);
    }
    #endif

    sceneColor          = drawbufferClamp(sceneColor);
}