#version 450 compatibility

/* RENDERTARGETS: 7,8 */
layout(location = 0) out vec3 fogScattering;
layout(location = 1) out vec3 fogTransmittance;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

uniform sampler2D colortex7, colortex8;
uniform sampler2D depthtex0;

uniform int frameCounter;

uniform float far, near;

uniform vec2 pixelSize, viewSize;

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"

/*
    Based on the bilateral blur chocapic13 does.
    Actually seems illegally more effective than my checkerboard solution. Can't make this shit up...
*/

#define FOG_SCATTER_IMAGE colortex7
#define FOG_TRANSMITTANCE_IMAGE colortex8

void BilateralFog(out vec3 scattering, out vec3 transmittance) {
    scattering  = vec3(0);
    transmittance = vec3(0);

    vec3 zmult      = vec3(1.0/(far * near),far+near,far-near);

    ivec2 depthUV   = ivec2(coord * viewSize);
    ivec2 fogUV     = ivec2(coord * viewSize / FOG_LOWRES_FACTOR);
    ivec2 jitter    = (ivec2(gl_FragCoord.xy + frameCounter) % 2) * 2;

    float depthCenter = depthLinear(texelFetch(depthtex0, depthUV, 0).x);

    float sumWeight = 0.0;

    ivec2 depthTapUV = depthUV + ivec2(-2) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
    float depthTap  = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
    float weight    = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
    ivec2 fogTapUV  = fogUV + ivec2(-2) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(-2, 0) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = fogUV + ivec2(-2, 0) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(0) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = fogUV + ivec2(0) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

        depthTapUV = depthUV + ivec2(0, -2) / FOG_LOWRES_FACTOR + jitter / FOG_LOWRES_FACTOR;
        depthTap   = depthLinear(texelFetch(depthtex0, depthTapUV, 0).x);
        weight     = abs(depthTap - depthCenter) < zmult.x ? 1.0 : 1e-5;
        fogTapUV   = fogUV + ivec2(0, -2) + jitter;
    scattering     += texelFetch(FOG_SCATTER_IMAGE, fogTapUV, 0).rgb * weight;
    transmittance  += texelFetch(FOG_TRANSMITTANCE_IMAGE, fogTapUV, 0).rgb * weight;
    sumWeight      += weight;

    scattering     /= sumWeight;
    transmittance  /= sumWeight;

    return;
}

void main() {
    BilateralFog(fogScattering, fogTransmittance);
}